#ifndef __TRANS_H__
#define __TRANS_H__
/*
 *   $RCSfile: trans.h,v $  
 *   $Revision: 2.10 $  
 *   $Date: 1993/04/21 17:26:06 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "trans_states.h"
#include "forward.h"

/*
 *	define a list structure for searching for the transaction id
 */
class SERVERTRANSREC {
public:
	TID				tid;
	UONE			serverTransState;	
	SERVERINFO		*serverInfo;	
	TRANSREC		*transRec; /* back ptr to the trans */
	LISTELEMENT		participants; /* for list of  serverinfos in transrec */
	LISTELEMENT		transList; /* for list of transrecs in serverinfo */

	void Init();
	void ReInit();
	LISTELEMENT *listlocation(int unique);
}; /* SERVERTRANSREC */


/*
 *	define the transaction control block
 */
struct	TransRec_s {

	LIST		userDescList;
	LIST		pageHashList;
	LIST		bufGroupList;
	LIST		lockHeaderList;
	LIST		participants; /* list of participating server/tid/states */
	int			numParticipants;
	UONE		clientTransState;
	BOOL		abortIfError; /* TRUE ==> sm_* aborts if any error occurs  */
	TID			clientTid;
	SERVERINFO	*coordInfo;
	TID			coordTid;

#if MAGIC_CHECKING IS_ENABLED

	MAGIC		magic;

#endif

};

#define CLIENT_ABORT_TID(tidptr) {\
	TRANSREC *transRec = findTransRec(tidptr);\
	if (transRec->clientTransState == T_ACTIVE) {\
		transRec->clientTransState = T_ABORT;\
	}\
}

#define CLIENT_ABORT_TRANS(transRec)\
	if (transRec->clientTransState == T_ACTIVE) {\
		transRec->clientTransState = T_ABORT;\
	}


/*
 *	define the transrec structure magic number
 */
#define TRANSREC_MAGIC	0x2ad505f0


/*
 *	define the magic checking macros
 */
#if MAGIC_CHECKING IS_ENABLED


#define INIT_TRANSREC_MAGIC(_transRec)				\
													\
	(_transRec)->magic = TRANSREC_MAGIC;

#define CHECK_TRANSREC_MAGIC(_transRec)				\
													\
	if ((_transRec)->magic != TRANSREC_MAGIC)	{	\
													\
		SM_ERROR(TYPE_FATAL, esmINTERNAL);			\
	}


#else


#define INIT_TRANSREC_MAGIC(_transRec)

#define CHECK_TRANSREC_MAGIC(_transRec)


#endif
#endif __TRANS_H__
