#ifndef __IO_VolRec_C__
#define __IO_VolRec_C__
/*
 *   $RCSfile: io_VolRec.C,v $  
 *   $Revision: 1.17 $  
 *   $Date: 1993/06/02 13:40:03 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

/*
 *****************************************************************
 *       c-callable wrapper funcs for page hash table and pool
 *****************************************************************
 */


#include "HashTable.h"
#include "Pool.h"
#include "resources.h"
#include "volume.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "lock.h"
#include "msgdefs.h"
#include "trans.h"
#include "io_extfuncs.h"
#include "io_intfuncs.h"
#include "io_globals.h"
#include "msg_funcs.h"
#include "sm_globals.h"
#include "trans.h"
#include "serverinfo.h"
#include "sm_state.h"
#include "trans_funcs.h"
#include "trans_globals.h"
#include "VOLREC.h"
#include "class_alias_macros.h"


HashTable<VOLIDCLASS,VOLREC>   	*volRecTable;
Pool<VOLREC>						*volRecPool;

static BOOL io_initialized = FALSE;


static struct { /* used in IO_FindVolId */
	VOLID 	volid;
	VOLREC 	*ptr;
} cached = { 0, NULL };


 int
io_Init ()
{
	if(io_initialized)
		return esmNOERROR;

	TRACE(TR_IO, TR_LEVEL_1);

	volRecPool = new Pool<VOLREC> ( "volRec", VOLREC_POOL,
		VolumeTableSize*5, POOL_NOMORE_GETMORE );

	if (volRecPool == NULL) {
		SM_ERROR(TYPE_SYS, esmMALLOCFAILED);
		return esmFAILURE;
	}
	/*
	volRecPool->Stats(stderr);
	*/

	volRecTable = new HashTable<VOLIDCLASS,VOLREC>( "volRec", 
		VOLREC_TABLE, VolumeTableSize);
	if (volRecTable == NULL) {
		SM_ERROR(TYPE_SYS, esmMALLOCFAILED);
		return esmFAILURE;
	}
	/*
	volRecTable->Stats(stderr);
	volRecTable->Dump(stderr, __LINE__,__FILE__);
	*/
	io_initialized = TRUE;
	return esmNOERROR;
}

#ifdef JUNK
VOLREC *
IO_NewVolRec()
{
	VOLREC *volRec;

	volRec = volRecPool->Get();
	return volRec;
}
#endif JUNK


/* 
 * IO_Mount: called from IO_verifyServer (called from io_FindVolId()) 
 * called ONLY if the volume is not already mounted.
 */
int
IO_Mount( 
	VOLREC *volRec
)
{
	MESSAGE			message;

	SM_ASSERT(LEVEL_1, ((volRec->flags & VOL_MOUNTED) == 0));

	SM_ASSERT(LEVEL_1, (volRec->serverInfo != NULL));
	if(volRec->serverInfo == NULL)
		return esmFAILURE;

	TRPRINT(TR_IO, TR_LEVEL_1, 
		("mounting volid:%d on server %s", 
		volRec->volid, volRec->serverInfo->hostName));

	message.header.type = MOUNT_VOLUME;
	message.header.replyRequested = TRUE;
	message.body.mount.volid = volRec->volid;
	/* message.header.magic is set in callServer */
	/* message.header.params.in not used */

	if (callServer(volRec->serverInfo, &message, NULL, 0) < 0)	{
		return(esmFAILURE);
    } 

	SM_ASSERT( LEVEL_1,  (message.body.mount.volid == volRec->volid) ); 

	/* we need to know of this is a temp volume */
	volRec->properties = message.body.mount.volumeProperties & VOLPROP_TEMP;

	if (receive(volRec->serverInfo, volRec->volName, DEVNAMEMAX) < 0){
		return esmFAILURE;
	}
	TRPRINT(TR_IO, TR_LEVEL_2, ("opened volume:%d(%s)", volRec->volid,
		volRec->volName));

	volRec->flags |= VOL_CLIENT_MOUNTED | VOL_MOUNTED;
	volRec->serverInfo->nMounted ++;
	return esmNOERROR;
} /* IO_Mount */

static BOOL
helperfunc1(
	VOLREC 	*volRec,
	char	*name,
	VOLREC 	**result
)
{
	int n = strlen(name);
	if( n>DEVNAMEMAX ) n = DEVNAMEMAX;

	if (!strncmp(name, volRec->volName, n))	{
		/* match */
		*result = volRec;
		return TRUE;
	}
	return FALSE;
}

void
IO_Dismount(
	VOLREC *volRec
)
{
	MESSAGE			message;

	TRACE(TR_IO, TR_LEVEL_1);

	if( !(volRec->flags & VOL_MOUNTED) ) {
		return; 
	} 
	SM_ASSERT(LEVEL_1,(volRec->serverInfo->flags & SERV_ESTAB));
		
	message.header.type = DISMOUNT_VOLUME;
	message.header.replyRequested = TRUE;
	/* message.header.magic is set in callServer */
	message.body.mount.volid = volRec->volid;

	callServer(volRec->serverInfo, &message, NULL, 0);
}

static BOOL
vol_Dismount (
	register VOLREC	*volRec,
	BOOL	serverAlso,
	void *serverInfo
)
{
	BOOL last4Server = FALSE;

	if((volRec->flags & VOL_MOUNTED)==0)
		return FALSE;

	/* non-null serverInfo means we want this to apply ONLY to that
	 * one server; furthermore, return TRUE when we have 
	 * dismounted its last volume 
	 */
	if(serverInfo && (serverInfo != (void *)(volRec->serverInfo)) )
		return FALSE;

	/* serverAlso argument tells if we want to send a message to 
	 * the server to have it dismount the volume too
	 */
	if(serverAlso)
		IO_Dismount(volRec);

	volRec->flags &= ~VOL_MOUNTED;


	if(--(volRec->serverInfo->nMounted) == 0) 
		last4Server = TRUE;

	if(serverInfo != NULL)
		return last4Server;

	return FALSE;
}

 VOLREC
*IO_FindVolId (

	VOLID	volid,
	FLAGS	flags
)
{
	VOLREC				*volRec;

	TRPRINT(TR_IO, TR_LEVEL_1, ("volid:%d", volid));
	if(volid == cached.volid) {
		volRec = cached.ptr;
	} else {
		CLASS_TYPE(VOLID)	*CLASS_ALIAS(volid) = (CLASS_TYPE(VOLID) *) &volid;

		volRec = volRecTable->Find(CLASS_ALIAS(volid));
	}

	if(volRec == NULL) {
		SM_ERROR(TYPE_USER, esmNOSUCHVOLUME);
		return NULL;
	}

	if(IO_verifyServer(flags, volRec, &TransRec) != esmNOERROR) {
		/* could be sm_errno==ECONNREFUSED, for example */
		return NULL;
	}

	cached.volid = volid;
	cached.ptr 	 = volRec;
	return volRec;
}


 int
io_Dismount (

	VOLID			volid 
)
{
	register VOLREC	*volRec;

	TRPRINT(TR_IO, TR_LEVEL_1, ("closing volume:%d", volid));

	volRec = IO_FindVolId(volid, SERV_MOUNTED_VOL);
	if (volRec ==  NULL)	{
		SM_ERROR(TYPE_USER, esmBADVOLID);
		return esmFAILURE;
	}
	(void) vol_Dismount(volRec, TRUE, NULL);
	return esmNOERROR;
}

 void
io_Shutdown ()

{
	TRACE(TR_IO, TR_LEVEL_1);

	cached.volid = 0;
	cached.ptr = NULL;
	delete volRecTable;
	volRecTable = NULL;
	delete volRecPool;
	volRecPool = NULL;
	io_initialized = FALSE;
}

 void
ForEachVolume(
	int		nargs,
	void 	*func, 
	void 	*arg1,
	void 	*arg2,
	void 	*arg3,
	void 	*arg4,
	void 	*arg5,
	void 	*arg6
)
{
	TRACE(TR_IO, TR_LEVEL_1);
	SM_ASSERT(LEVEL_1, (nargs <= 7));
	SM_ASSERT(LEVEL_1, (nargs >= 1));
#ifdef __GNUC__
	volRecTable->ForEach(nargs, func, arg1, arg2, arg3, arg4, arg5, arg6);
#else __GNUC__
	volRecTable->ForEach(nargs, (FOREACHFUNC)func, 
		arg1, arg2, arg3, arg4, arg5, arg6);
#endif __GNUC__
}

void
serverDied(void *serverInfo)
{
	TRACE(TR_IO, TR_LEVEL_1);

#ifdef __GNUC__
	volRecTable->ForEach(3, vol_Dismount, FALSE, serverInfo);
#else __GNUC__
	volRecTable->ForEach(3, (FOREACHFUNC)vol_Dismount, FALSE, serverInfo);
#endif __GNUC__
}

 void
io_Final ()
{
	TRACE(TR_IO, TR_LEVEL_1);

#ifdef __GNUC__
	volRecTable->ForEach(3, vol_Dismount, TRUE, NULL);
#else __GNUC__
	volRecTable->ForEach(3, (FOREACHFUNC)vol_Dismount, TRUE, NULL);
#endif __GNUC__
}

int
IO_verifyServer(
	FLAGS 		flags,
	VOLREC 		*volRec, 
	TRANSREC	*transRec
)
{
	SERVERINFO			*serverInfo = volRec->serverInfo;
	SERVERTRANSREC 		*st = NULL;

	SM_ASSERT(LEVEL_1, (serverInfo != NULL));
	/*
	 * be sure that we're asking something meaningful
	 */
	SM_ASSERT(LEVEL_1, ((flags & (SERV_MOUNTED_VOL|SERV_IN_TX|SERV_ESTAB))!=0));

	if(flags & SERV_IN_TX) {
		flags |= SERV_ESTAB;

		SM_ASSERT(LEVEL_1, (transRec != NULL));
		TRPRINT(TR_TRANS, TR_LEVEL_1, 
			("searching for server in participants"));
		for (st = (SERVERTRANSREC *)
			FIRST_LIST_ELEMENT(&(transRec->participants));
			(st != NULL) && (st->serverInfo != serverInfo);
			st = (SERVERTRANSREC *)NEXT_LIST_ELEMENT(&(st->participants)) 
			) ;

		if( st == NULL ) {
			/* server is not (yet) used in this tx - add it to the list */
			TRPRINT(TR_TRANS, TR_LEVEL_1, ("%s is new participant",
				serverInfo->hostName));
			st = addTransServerInfo(transRec, serverInfo);
			SM_ASSERT(LEVEL_1, (st->serverInfo == serverInfo));
			SM_ASSERT(LEVEL_1, (st->transRec == transRec));
			SM_ASSERT(LEVEL_1, (st->serverTransState==T_INACTIVE));
		} 
		TRPRINT(TR_TRANS, TR_LEVEL_1, ("%s tx state is %d",
					serverInfo->hostName, st->serverTransState));
		switch(st->serverTransState) {
		case T_INACTIVE: 
			break;

		case T_PREPARED: 
			fprintf(stderr, "%d %s WHAT TODO HERE?",__LINE__,__FILE__);
			break;

		case T_ACTIVE: 
			SM_ASSERT(LEVEL_1, (serverInfo->flags & SERV_ESTAB) );
			st = NULL; /* so we don't do a begin_tx below */
			break;

		case T_ABORT: 
			SM_ERROR(TYPE_USER, esmTRANSABORTED);
			return  esmFAILURE;

		default:
			SM_ERROR(TYPE_USER, esmINTERNAL);
		}
		SM_ASSERT(LEVEL_1, ((st == NULL) || (st->serverTransState==T_INACTIVE)));
	}
	if(flags & (SERV_ESTAB | SERV_MOUNTED_VOL)) {
		if( (st != NULL) &&
			(serverInfo->flags & SERV_ESTAB) ) {
			/* it was once established but we have since 
			 * completed the transaction, so the server could have
			 * timed us out.  So we check to see if we are still
			 * connected before we do the begin-transaction
			 */
			SM_ASSERT(LEVEL_1, (st->serverTransState==T_INACTIVE));

			struct	sockaddr name;
			int				 namelen = sizeof(name);

			TRPRINT(TR_TRANS, TR_LEVEL_1, ("%s INACTIVE: checking for cx",
				serverInfo->hostName));

			if(  getpeername(serverInfo->socket, &name, &namelen)  < 0) {
				if(errno == ENOTCONN) {
					serverInfo->flags = 
			SERV_NOTCONNECTED | (serverInfo->flags & SERV_CONTACTED_IN_PAST);
				} else {
					SM_ERROR(TYPE_SYS, errno);
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}
			}
		}
		if( !(serverInfo->flags & SERV_ESTAB) ) {
			if(	server_connectInit(volRec->serverInfo, 
							FALSE /* not admin client */) != esmNOERROR) 	{
				SM_ERROR(TYPE_SYS, sm_errno);
				return esmFAILURE;
			}
		}
	}
	if(st != NULL) {
		SM_ASSERT(LEVEL_1, (st->serverTransState==T_INACTIVE));
		if(server_beginTrans(serverInfo, &(st->tid))==esmNOERROR){
			 st->serverTransState = T_ACTIVE ;
		} else  {
			return esmFAILURE;
		}
	}
	if(flags & SERV_MOUNTED_VOL) {
		if( ! (volRec->flags & VOL_MOUNTED) ) {
			if( IO_Mount(volRec) != esmNOERROR) {
				SM_ERROR(TYPE_USER, sm_errno);
				return esmFAILURE;
			}
		}
	}
	return esmNOERROR;
}

 VOLREC
*io_FindVolId (

	VOLID	volid 
)
{
	return IO_FindVolId (volid, SERV_MOUNTED_VOL|SERV_IN_TX);
}


 int
io_AddVolume (
	VOLID			volid,
	char			*host,
	char			*port
)
{
	VOLREC				*volRec;
	CLASS_TYPE(VOLID)	*CLASS_ALIAS(volid) = (CLASS_TYPE(VOLID) *)&volid;

	if( io_Init() != esmNOERROR) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	volRec = volRecTable->Find(CLASS_ALIAS(volid));

	if(volRec == NULL) {
		volRec = volRecPool->Get();

		volRec->volid = volid;

		volRecTable->Insert(volRec);
	}
	volRec->serverInfo = getServerInfo(host, port);

	if (volRec->serverInfo == NULL) {
		return(esmFAILURE);
	}

	return esmNOERROR;
}

 VOLREC*
io_Mount (
	VOLID			volid
)
{
	VOLREC 	*volRec;

	volRec = IO_FindVolId(volid, SERV_MOUNTED_VOL);
	/* if it returns a volRec, the volume is mounted at the server */
	return volRec;
}

#ifdef DEBUG
extern "C" void DumpVolRecTable();

void
DumpVolRecTable()
{
		volRecTable->Dump(stderr, __LINE__,__FILE__);
}
#endif DEBUG
#endif __IO_VolRec_C__
