/*
 *   $RCSfile: lg_PrintIndex.c,v $  
 *   $Revision: 2.7 $  
 *   $Date: 1993/04/13 17:38:40 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "lock.h"
#include "chunk.h"
#include "volume.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "io_extfuncs.h"
#include "error.h"

#ifdef DEBUG


static SHORTPID	PidArray[10000];
static int		PidCount;

#ifdef JUNK
	static VOLREC	*diskRec;
#endif JUNK


/*
 * Routine to print the contents of a large object index for debugging
 * purposes.
 */



 void 
lg_PrintIndex (

	BUFGROUP		*bufGroup,	/* buffer group for accesses	*/
	PID				*nodePid,	/* root pid for index subtree	*/
	int				indent,
	BOOL 			flag,
	SLOTINDEX		rootSlot 
)
{

    register int	i;			/* used for loop index			*/
    LGNODE          *nodePtr;	/* pointer to current node		*/
	GROUPLINK		*nodeLink;
	int				j;

	TRACE(TR_LG, TR_LEVEL_1);

	/*
	 *	Check for a flags
	 */
	if (flag)	{
		
		/*
		 *	Initialize the number
		 */
		PidCount = 0;

#ifdef JUNK
		diskRec = io_GetVolume(nodePid->Pvolid);
#endif
	}
		
    /*
     * Read the node and then print its contents out.  Indent an amount
     * proportional to the level so the output will be readable.
     */
    if ((nodeLink = bf_ReadPage(bufGroup, nodePid,
		(rootSlot != NIL ? H_SLOTTED : H_LARGE),
		(rootSlot != NIL ? SLOTTED_PAGE2SIZE : LG_PAGE2SIZE), NL)) == NULL)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	check to see if this is a slotted root
	 */
	if (rootSlot != NIL)	{

		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*rootPtr;


		TRPRINT(TR_LG, TR_LEVEL_3, ("slot root pid:%d slot:%d", nodePid->page, rootSlot));

		/*
		 *	Get a pointer to the page, slot and object
		 */
		pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, rootSlot);
		rootPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

		/*
		 *	get a pointer to the node
		 */
		nodePtr = (LGNODE *) &(rootPtr->rootPage);

	} else {

		/*
		 *	get a pointer to the node
		 */
		nodePtr = (LGNODE *) nodeLink->bufFrame;
	}

	/*
	 *	print the indenting tabs
	 */
	for (i = 0; i < indent; i++)	{

		printf("\t");
	}

	/*
	 *	print the node contents
	 */
    lg_PrintNode(nodePtr, indent, nodePid->page);

    /*
     * If we're not at the parent of a leaf, then run print each subtree
     * rooted here;  otherwise, we're all done.
     */
    if (nodePtr->header.height > 1) {

#ifdef JUNK
		/*
		 *	check the bitmap to see if it is set
		 */
		if (checkset(diskRec->freeBitmap, nodePid->page))	{

			SM_ERROR(TYPE_FATAL, -1);
		}
#endif JUNK

		/*
		 *	check to see if the pid is in the list
		 */
		for (j = 0; j < PidCount; j++)	{

			/*
			 *	check for a pid match
			 */
			if (PidArray[j] == nodePid->page)	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
		}

		/*
		 *	put the pid in the array
		 */
		PidArray[PidCount] = nodePid->page;

		/*
		 *	print all subTrees
		 */
        for (i = 0; i < nodePtr->header.numSlots; i++) {

            lg_PrintIndex(bufGroup, &(nodePtr->slot[i].pid), indent + 1, FALSE, NIL);
        }

    } else {

		/*
		 *	run through the leaf nodes
		 */
		for (i = 0; i < nodePtr->header.numSlots; i++)	{

#ifdef JUNK
			/*
			 *	check the bitmap to see if it is set
			 */
			if (checkset(diskRec->freeBitmap, nodePtr->slot[i].pid.page) ||
				checkset(diskRec->freeBitmap, nodePtr->slot[i].pid.page + 1))	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
#endif JUNK

			/*
			 *	check for a match in the pid array
			 */
			for (j = 0; j < PidCount; j++)	{

				/*
				 *	check for match
				 */
				if (PidArray[j] == nodePtr->slot[i].pid.page)	{

					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}
			}

			/*
			 *	stick the pid in the array
			 */
			PidArray[PidCount] = nodePtr->slot[i].pid.page;
			PidCount++;
		}
	}

    /*
     * Lastly, unfix the page.
     */
    bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
}

#endif
