/*
 *   $RCSfile: sm_CreateVersion.c,v $  
 *   $Revision: 2.24 $  
 *   $Date: 1993/03/23 23:41:56 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "trans.h"
#include "pool.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "volume.h"
#include "file.h"
#include "lgobject.h"
#include "lsn.h"
#include "loginfo.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "version_graph.h"
#include "version_funcs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "fi_extfuncs.h"
#include "lg_versionfuncs.h"
#include "sm_intfuncs.h"
#include "sm_version_intfuncs.h"
#include "sm_extfuncs.h"
#include "clog_extfuncs.h"
#include "util_funcs.h"
#include "trans_funcs.h"
#include "bf_globals.h"
#include "trans_globals.h"
#include "clog_globals.h"


 int
sm_CreateVersion (
	int				groupIndex,
	int				nearType,
	OID				*parentObj,
	OID				*nearObj,
	OID				*oid 
)
{
    PAGESLOT		*slotPtr;		/* slot pointer					*/
    LARGEOBJHDR		*largeObject;	/* ptr to large object header	*/
    SMALLOBJ		*smallObject;	/* small object pointer			*/
    SLOTTEDPAGE		*homePage= NULL;	/* ptr to home page				*/
    SLOTTEDPAGE		*parentPage;	/* ptr to parent page			*/
    SLOTINDEX		emptySlot;		/* slot used for the new obj	*/
    int				spaceNeeded;	/* space needed for the new obj */
	int				alignLength;
    BOOL			creatingLargeObj;/* whether creating large obj	*/
    BOOL			tryNearPage;	/* whether to look at near page */
    BOOL			allocatePage;	/* whether to allocate new page */
    BOOL			setNeighborLink;/* whether to set neighbor pid link */
    BOOL			allocateSlot;	/* whether we need to allocate slot */
	BOOL			rootObject;
    PID				nearPid;		/* create the obj near this page */
    SLOTTEDPAGE		*nearPage;		/* indirect ptr to near page	*/
    PID				homePid;		/* the new obj's home page		*/
    PID				rootPid;		/* large object root page		*/
    PID				parentPid;		/* parent object header page 	*/ 
	GROUPLINK		*nearLink = NULL;
	BOOL			nearLinkIsDirty = FALSE;
	GROUPLINK		*homeLink= NULL;
	BOOL			homeLinkIsDirty = FALSE;
	GROUPLINK		*parentLink= NULL;
	BUFGROUP		*bufGroup;
	UNIQUE			unique;
	OBJECT		  	*parentObjPtr;	 /* ptr to parent object which is a small object */	
	SMALLOBJ  		*smallParentObjPtr; /* ptr to parent object which is a small object */	
	LARGEOBJHDR    	*lgParentObjPtr;/* ptr to parent object    		*/	
	SLOTINDEX		objectSlot;
	FID				fid;
	VERSIONGRAPH	*vhg;			/* version history graph 		*/
	VHGNODEPTR		vhgPtr;			/* "pointer" to node in VHG 	*/
	OPERATIONINFO	opInfo;			/* operation info used for logging*/
	int				smDataSize;     /* data portion size for lg obj hdr*/

	TRPRINT(TR_SM, TR_LEVEL_1, ("Create version"));

	INIT_SMSTATE(&SM_State);

	/* check to see if a transaction is active */
	if (checkTransState(parentObj->diskAddr.volid))	{
		return esmFAILURE;
	}

	bufGroup = &BufGroupPtr[groupIndex];
	if (CHECK_BUFGROUP_VALID(bufGroup))	{
		SM_ERROR(TYPE_USER, esmBADGROUP);
		return esmFAILURE;
	}

	/* Construct the parent pid */
	parentPid.volid = parentObj->diskAddr.volid; 
	parentPid.page = parentObj->diskAddr.page; 

    /*
     *  read the parent page 
     */
    if ((parentLink = bf_ReadPage(bufGroup, &parentPid, H_SLOTTED,
                                SLOTTED_PAGE2SIZE, SH)) == NULL)    {
		FAIL
    }

	if (SM_SetLogServerState(parentLink, NULL) != esmNOERROR) {
		FAIL
	}
    parentPage = (SLOTTEDPAGE *) parentLink->bufFrame;

    if (CHECK_SLOTTED_MAGIC(parentPage))  {
        SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
    }

	/* Get the file id for the page */
	fid = parentPage->header.fid;

    objectSlot = parentObj->diskAddr.slot;
    slotPtr = GETSLOTPTR(parentPage, objectSlot);

    /* check the validity of the oid */
    if (CHECKOID(objectSlot, parentPage, slotPtr, parentObj->diskAddr.unique))  {
        SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
    }

    /* get a pointer to the object */
    parentObjPtr = GETOBJECTPTR(parentPage, slotPtr);

	/**********************************************************/

	/*
	 *	Validate that the parent can have a version made.   It must 
	 *	be both versioned and frozen.
	 */
	if (parentObjPtr->objectHdr.properties & P_VERSIONED) {
		if ( !(parentObjPtr->objectHdr.properties & P_FROZEN) ) {
			SM_ERROR(TYPE_USER, esmNOTFROZEN);
			FAIL
		}
	} else {
        SM_ERROR(TYPE_USER, esmNOTVERSIONED);
		FAIL
	}

	/*
	 *	Make sure object is not the forwarded record of a moved
	 *	object.
	 */
	if (parentObjPtr->objectHdr.properties & P_FORWARD) {
        SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

	/*
	 *	Are we creating a large or small object
	 */
	if (parentObjPtr->objectHdr.properties & P_LARGEOBJ) {
		creatingLargeObj = TRUE;
		lgParentObjPtr = (LARGEOBJHDR*) parentObjPtr;

		/* Get a "pointer" the the version history graph for the obj */
		if (lg_GetVHGPtr(bufGroup, lgParentObjPtr, 
						 parentObj->diskAddr.volid, &vhgPtr)) {
			FAIL
		}

		/* Determine how much space is needed */
		if (lgParentObjPtr->objectHdr.properties & P_HEADERONPAGE) {

			CHECK_SLOTROOT_MAGIC( (&(((LARGEOBJROOT*)lgParentObjPtr)->rootPage)) );
			/* This object has the root on this slotted page */
			rootObject  = TRUE;

			/* allocate space for the header and the root */
			spaceNeeded = ALIGNSIZE(sizeof(LARGEOBJROOT)) + sizeof(PAGESLOT);
			TRPRINT(TR_SM, TR_LEVEL_2, ("creating slotted large version"));
		} else {
			/* This object has the root on a separate page */
			rootObject  = FALSE;

			/* allocate enough space to hold the header only */
			spaceNeeded = sizeof(LARGEOBJHDR) + sizeof(PAGESLOT);
			TRPRINT(TR_SM, TR_LEVEL_2, ("creating large object version on seperate"));
		}

		/* Make sure parent object has a version history graph oid */
		if (vhgPtr.oid.diskAddr.volid == 0) {	
			TRPRINT(TR_SM, TR_LEVEL_2, ("Need to create VHG"));
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	} else {

		/*
		 *	mark the fact that we are creating a small object version
		 */
		creatingLargeObj = FALSE;
		smallParentObjPtr = (SMALLOBJ*) parentObjPtr;
		alignLength = ALIGNSIZE((int)smallParentObjPtr->objectHdr.size);
		spaceNeeded		 = sizeof(OBJHDR) + alignLength + sizeof(PAGESLOT);
		TRPRINT(TR_SM, TR_LEVEL_2, ("creating small object version"));
	}

	TRPRINT(TR_SM, TR_LEVEL_2, ("spaced needed:%d", spaceNeeded));

	/*
	 *  If the version is a large object, get the version history
	 *	graph now since it's possible that the graph lies on the
	 *	page where the create will occur.  If the VHG is not obtained
	 * 	now, then an error will occur if it does lie on the
	 * 	same page, since the following code changes slot info for
	 * 	the page.
	 */
	vhg = NULL;
	if (creatingLargeObj == TRUE) {
		/*
		 *  Get the version history graph, making sure there is room
		 *	to add a node.  Remember to free the space for the VHG
		 *	on an error, or before exiting.
		 */
		if (SM_GetVersionGraph(bufGroup, &vhgPtr.oid, &vhg, TRUE, EX)) {
			FAIL
		}
 	}

    /*
     *	Switch on near to determine where we should
     *	try and put the new object.
     */
	if (SM_DecideNear(nearType, nearObj, &nearPid, &fid, &tryNearPage))	{
		FAIL
	}

    /*
     *	Try and create the new object on nearObj's home page.
     */
    if (tryNearPage) {

		TRPRINT(TR_SM, TR_LEVEL_2, ("attempting to create on near page"));

		/*
		 *	call the routine to find a near page that
		 *	can be used
		 */
		if (SM_TryNearPage(bufGroup, &fid, spaceNeeded, &allocatePage,
						   &nearPid, &nearLink, &tryNearPage))	{
			FAIL
		}

		/* check to see if we have a valid page */
		if (!allocatePage)	{
			/* setup the home fields */
			homePid = nearPid;
			homeLink = nearLink;
			homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
			nearLink = NULL; /* so we don't unfix twice */
		} else if (tryNearPage)	{
			/* just set nearPage for future use */
			nearPage = (SLOTTEDPAGE *) nearLink->bufFrame;
		}
    } else /* tryNearPage == FALSE */ {
		TRPRINT(TR_SM, TR_LEVEL_2, ("not using near page"));
		/*
		 * In this case, there is no nearObj page to look at and
		 * we have to allocate a page.
		 */
		allocatePage = TRUE;
    }

    /*
     * Figure out whether we have to set the neighbor pid link.
     */
    setNeighborLink = tryNearPage && allocatePage;

    /*
     *	Allocate a page if necessary.
     *	If setNeighborLink is TRUE, then nearPage is valid
     *	and points to the near page whose neighbor pid link has to be set.
     */
    if (allocatePage == TRUE) {
		/*
		 * Do the low-level allocation.
		 */
		if (io_AllocSlottedPages(1, &nearPid, &homePid, &fid, creatingLargeObj,
								 &unique, bufGroup))	{
			FAIL
		}
		TRPRINT(TR_SM, TR_LEVEL_2, ("allocated new page:%d", homePid.page));

		sm_SetAbortIfError();

		/*
		 * Get a buffer for the page.
		 */
		if ((homeLink = bf_FakeReadPage(bufGroup, &homePid, H_SLOTTED,
								  SLOTTED_PAGE2SIZE)) == NULL)	{
			FAIL
		}
		homePage = (SLOTTEDPAGE *) homeLink->bufFrame;

		/* Initialize the page.  */
		INITPAGE(homePage, homePid, fid, creatingLargeObj, unique);
		homePage->header.lrc = *(GET_LOGINFO->initLRC);
		homeLinkIsDirty = TRUE;

		/* Set the neighbor pid link if necessary.  */
		if (setNeighborLink == TRUE) {

			TRPRINT(TR_SM, TR_LEVEL_2, ("setting neighbor link"));

			/*
			 *	set the link and but don't dirty the page since it
			 *	is just a hint
			 */
			nearPage->header.neighborPid = homePid;
			nearLinkIsDirty = TRUE;
			TRPRINT(TR_SM, TR_LEVEL_2, ("setting neighbor link"));
		}
    } else {
		/*
		 *	Check to see if we are creating a large object
		 */
		if (creatingLargeObj == TRUE)	{

			/*
			 *  Record the fact that a large object has been created on 
			 *  the page.
			 */
			if (SM_LargeObjOnPage(homePage)) {
				FAIL
			}
		}
    }
	homeLinkIsDirty = TRUE;
	sm_SetAbortIfError();

	/*
	 *	check to see if there are any stray free slots
	 */
	if (homePage->header.freeSlot == NIL)	{
		TRPRINT(TR_SM, TR_LEVEL_2, ("slot is new"));

		/* Just grab the next slot after current last */
		emptySlot = homePage->header.slotCount;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));

		slotPtr = GETSLOTPTR(homePage, emptySlot);

		/* increment the number of slots in use */
		homePage->header.slotCount++;
		allocateSlot = TRUE;
	} else {
		TRPRINT(TR_SM, TR_LEVEL_2, ("old slot is free"));

		/* take the stray free slot */
		emptySlot = homePage->header.freeSlot;
		TRPRINT(TR_SM, TR_LEVEL_2, ("using free slot:%d", emptySlot));
		slotPtr = GETSLOTPTR(homePage, emptySlot);

		/* get a new head of the list */
		homePage->header.freeSlot = (SLOTINDEX) slotPtr->unique;
		TRPRINT(TR_SM, TR_LEVEL_2, ("new free slot:%d", homePage->header.freeSlot));
		allocateSlot = FALSE;
	}

    /*
     * Now create the appropriate type of object.
     */
	if (creatingLargeObj == TRUE) {

		VHGNODEID		newNodeId;		/* new VHG node		    */

		/*
		 *	fill in the root page in case we have the 
		 *	large object header on the slotted page
		 */
		rootPid = homePid;

		/* Update the home page info.  */
		slotPtr->offset = homePage->header.freeStart;

		/*
		 *	Make sure the version history graph has been obtained
		 *	Remember to free the space for the VHG on an error, 
		 *  or before exiting.
		 */
		SM_ASSERT(LEVEL_3, vhg != NULL)	;
	
		/* Add a new version to the VHG */
		if (createVersion(vhg, parentObj, vhgPtr.nodeId, &newNodeId)) {
			FAIL
		}

		/*
		 *	Copy the large obj header to the page for the new version
		 */
		largeObject = (LARGEOBJHDR *) ((char *) homePage + homePage->header.freeStart);
		if (rootObject) {
			*(LARGEOBJROOT*)largeObject = *(LARGEOBJROOT*)lgParentObjPtr;
			CHECK_SLOTROOT_MAGIC( (&((LARGEOBJROOT*)largeObject)->rootPage) );
			/*
			 *	Record the nodeId in the VHG for this object
			 */
			((LARGEOBJROOT*)largeObject)->rootPage.header.vhgPtr.nodeId	= newNodeId;

		} else {
			/* Copy the header */
			*largeObject = *lgParentObjPtr;

			/* Create a new root page for the version */
			if (lg_CreateVersionRoot(bufGroup, &fid, &nearPid, 
									 largeObject, lgParentObjPtr, 
									 newNodeId)) {
				FAIL
			}

		}

		/* check to see if we can use the unique on the page */
		if (homePage->header.uniqueCount == 0)	{

			/* get a new unique range */
			if (io_GetUnique(&homePid, &unique, UNIQUE_ALLOC_COUNT))	{
				FAIL
			}

			/* set the unique fields */
			homePage->header.unique = unique;
			homePage->header.uniqueCount = UNIQUE_ALLOC_COUNT;
		}

		/* fill in the unique field */
		slotPtr->unique = homePage->header.unique++;
		homePage->header.uniqueCount--;
		TRPRINT(TR_SM, TR_LEVEL_2, ("unique:%x", slotPtr->unique));

		/* Check to see if we created a root object */
		if (rootObject)	{
			LGSLOT*	slot; /* array of slots on root */

			/* set the properties */
			largeObject->objectHdr.properties = P_LARGEOBJ|P_HEADERONPAGE|P_VERSIONED;

			/* update the page usage statistics */
			homePage->header.freeStart += ALIGNSIZE(sizeof(LARGEOBJROOT));
			homePage->header.freeBytes -= ALIGNSIZE(sizeof(LARGEOBJROOT));
			/* Mark the pointers in the large obj root as shared */
			for (slot = ((LARGEOBJROOT*)largeObject)->rootPage.slot;
				 slot < ((LARGEOBJROOT*)largeObject)->rootPage.slot+LGMAXPAGESLOTS;
				 slot++ ) {
				 slot->shared = TRUE;
			}
			/* calc the size of the data portion of the header - for logging */
			smDataSize = sizeof(LARGEOBJROOT) - sizeof(OBJHDR);
		} else {
			largeObject->objectHdr.properties = P_LARGEOBJ|P_VERSIONED;

			homePage->header.freeStart += sizeof(LARGEOBJHDR);
			homePage->header.freeBytes -= sizeof(LARGEOBJHDR);

			/* calc the size of the data portion of the header - for logging */
			smDataSize = sizeof(LARGEOBJHDR) - sizeof(OBJHDR);
		}

		/* if we are allocating a slot then subtract that space */
		if (allocateSlot == TRUE) {
			homePage->header.freeBytes -= sizeof(PAGESLOT);
		}

		/*
		 * Set the OID, dirty the home page, unfix it and return.
		 */
		oid->diskAddr.page	 = homePid.page;
		oid->diskAddr.slot	 = emptySlot;
		oid->diskAddr.unique = slotPtr->unique;
		oid->diskAddr.volid  = homePid.volid;
		TRPRINT(TR_SM, TR_LEVEL_2, ("returning oid pid:%d slot:%d unique:%x",
				oid->diskAddr.page, emptySlot, slotPtr->unique));

		if (GET_LOGLEVEL(LOG_SPACE)) {
			/* log the create of the large object header as a small object*/
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ;
			opInfo.flags = 0;
			opInfo.oid = *oid;
			opInfo.start = 0;
			opInfo.size = smDataSize;
			opInfo.oldData = NULL;
			opInfo.newData = (char *) largeObject;
			opInfo.actionPid.volid = oid->diskAddr.volid;
			opInfo.actionPid.page  = oid->diskAddr.page;
			if(clog_LogCreateObj(&opInfo) != esmNOERROR) {
				FAIL
			}
		}

		/* Store the oid for the version in the VHG node */
		vhg->nodeArray[newNodeId].oid = *oid;

		/*
		 *	Write the updated version history graph and free space
		 *	pointed to by vhg
		 */
		if (SM_SetVersionGraph(bufGroup, &vhgPtr.oid, vhg)) {
			FAIL
		}
		goto Success;
    }

    /*
	 *	If we got here, we're creating a small object. 
	 */
    smallObject = (SMALLOBJ *) (((char *) homePage) + homePage->header.freeStart);

	/* load the object header */
	smallObject->objectHdr = smallParentObjPtr->objectHdr;
    smallObject->objectHdr.properties = P_VERSIONED;
	TRPRINT(TR_SM, TR_LEVEL_2, ("freestart:%d length:%d",
			homePage->header.freeStart, smallObject->objectHdr.size));

#if OBJECT_MAGIC IS_ENABLED
	smallObject->objectHdr.magic = OBJECT_HDR_MAGIC;
#endif OBJECT_MAGIC IS_ENABLED

    /*
     *	Copy the data for the new version.  First we must see if the
	 *	object has been move so that we can get the data from the
	 *	new location.
     */
	if (smallParentObjPtr->objectHdr.properties & P_MOVED) {
		/*
		 *	The object has been moved, so get the forwarded object
		 */
		GROUPLINK       *forwardLink;
		PAGESLOT        *forwardSlotPtr;
		SLOTTEDPAGE     *forwardPage;
		int             forwardSlot;
		PAGEINFO   		*pageInfo;
		int				length;

		TRPRINT(TR_SM, TR_LEVEL_2, ("version of moved small obj"));

		/*
		 *  we must read in the forwarded oid
		 */
		length = -1;
		if (SM_ReadForward(bufGroup, &forwardLink, 
							(OID *) smallParentObjPtr->data, 
							0, &length)) {
			FAIL
		}
		
		/* Get a pointer to the page info structure */
		pageInfo = forwardLink->pageHash->pageInfo;

		/* get a pointer to the forwarded page */
		forwardPage = (SLOTTEDPAGE *) forwardLink->bufFrame;

		/* get a pointer to the object */
		forwardSlot = ((OID *) smallObject->data)->diskAddr.slot;
		forwardSlotPtr = GETSLOTPTR(forwardPage, forwardSlot);
		smallObject = (SMALLOBJ *) GETOBJECTPTR(forwardPage, forwardSlotPtr);

		/* Copy the data from the small object */
		bcopy(smallParentObjPtr->data, smallObject->data, 
			  (int)parentObjPtr->objectHdr.size);

	} else {
		/* Just copy the data from the small object */
		TRPRINT(TR_SM, TR_LEVEL_2, ("version of unmoved small obj"));
		bcopy(smallParentObjPtr->data, smallObject->data, 
			  (int)parentObjPtr->objectHdr.size);
	}

	/* Initialize the contents of the slot */
    slotPtr->offset = homePage->header.freeStart;

	/*
	 *	check to see if we can use the unique on the page
	 */
	if (homePage->header.uniqueCount == 0)	{

		/* get a new unique range */
		if (io_GetUnique(&homePid, &unique, UNIQUE_ALLOC_COUNT))	{
			FAIL
		}
		homePage->header.unique = unique;
		homePage->header.uniqueCount = UNIQUE_ALLOC_COUNT;
	}

	/* fill in the unique field */
	slotPtr->unique = homePage->header.unique++;
	homePage->header.uniqueCount--;
	TRPRINT(TR_SM, TR_LEVEL_2, ("unique:%x", slotPtr->unique));

	/*
     *	Update the home page stats to account for the
	 *	space allocated for the object
     */
    homePage->header.freeStart += sizeof(OBJHDR) + alignLength;
    homePage->header.freeBytes -= sizeof(OBJHDR) + alignLength;

	/* Update free space if we allocated a new slot */
    if (allocateSlot == TRUE) {
		homePage->header.freeBytes -= sizeof(PAGESLOT);
    }
	TRPRINT(TR_SM, TR_LEVEL_2, ("freeStart:%d freeBytes:%d",
			homePage->header.freeStart, homePage->header.freeBytes));

    /*
     * Fill in the oid for return to the user
     */
    oid->diskAddr.page	 = homePid.page;
    oid->diskAddr.slot	 = emptySlot;
	oid->diskAddr.unique = slotPtr->unique;
	oid->diskAddr.volid  = homePid.volid;
	TRPRINT(TR_SM, TR_LEVEL_2, ("oid page:%d slot:%d unique:%d",
			oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	if (GET_LOGLEVEL(LOG_SPACE)) {

		/* Fill in the operation info */
		opInfo.flags = 0;
		opInfo.oid = *oid;
		opInfo.start = 0;
		opInfo.size = (int)smallObject->objectHdr.size;
		opInfo.oldData = NULL;
		opInfo.newData = (char *) smallObject;
		opInfo.actionPid.volid = oid->diskAddr.volid;
		opInfo.actionPid.page  = oid->diskAddr.page;

		if (GET_LOGLEVEL(LOG_DATA)) {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ;
			/* Log the create operation */
			if(clog_LogCreateObj(&opInfo)) {
				FAIL
			}
		} else {
			opInfo.action = LOG_ACTION_CREATE_SMALL_OBJ_ND;
			/* Log the create operation */
			if(clog_LogCreateObjND(&opInfo)) {
				FAIL
			}
		}
	}

Success:
	if(homePage) {
		sm_AuditSlotPage(homePage);
	}
	if(nearLink) {
		bf_UnfixPage(nearLink, BF_DEFAULT, nearLinkIsDirty);
	}
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
	}
	if(parentLink) {
		bf_UnfixPage(parentLink, BF_DEFAULT, FALSE); /* always */
	}
	BF_AuditFreeList();
    return esmNOERROR;

FailureReturn:
	if(nearLink) {
		bf_UnfixPage(nearLink, BF_DEFAULT, nearLinkIsDirty);
	}
	if(homeLink) {
		bf_UnfixPage(homeLink, BF_DEFAULT, homeLinkIsDirty);
	}
	if(parentLink) {
		bf_UnfixPage(parentLink, BF_DEFAULT, FALSE); /* always */
	}
	SM_AbortIfRequired();
	return esmFAILURE;
}
