/*
 *   $RCSfile: sm_FetchInit.C,v $  
 *   $Revision: 1.12 $  
 *   $Date: 1993/04/08 16:19:05 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "io.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "volume.h"
#include "bf_globals.h"
#include "bf_extfuncs.h"
#include "trans.h"
#include "trans_globals.h"
#include "load.h"
#include "scan.h"
#include "sm_params.h"
#include "btree.h"
#include "sm_extfuncs.h"
#include "sm_intfuncs.h"
#include "bt_extfuncs.h"
#include "lh_extfuncs.h"
#include "trans_funcs.h"
#include "sm_state.h"


int sm_FetchInit(IID* ndx, int groupIndex, KEY* bound1, SMCOND cond1,
						KEY* bound2, SMCOND cond2, SMCURSOR* cursor)
{
	BUFGROUP*	bufGroup = & BufGroupPtr[groupIndex];
	GROUPLINK*	gLink;

    if (CHECK_BUFGROUP_VALID(bufGroup))  {
        SM_ERROR(TYPE_USER, esmBADGROUP);
        return esmFAILURE;
    }

    INIT_SMSTATE(&SM_State);

    int loadNdx;
    if (sm_FindIndexLoad(ndx, &loadNdx))  {
        return esmFAILURE;
        }
    if (loadNdx >= 0)  {
        SM_ERROR(TYPE_USER, esmBULKLOADING);
        return esmFAILURE;
        }

    // check to see if a transaction is active
    if (checkTransState(ndx->pid.volid))  {
        return(esmFAILURE);
    }
	
	/*
	 *	Read in the index descriptor page
	 */
	if ((gLink = bf_ReadPage(bufGroup, &ndx->pid, H_INDEXDESC, 
										INDEXDESC_PAGE2SIZE, SH)) == NULL) 
		return esmFAILURE;
	
	SM_ASSERT(LEVEL_3, ((INDEXDESCPAGE*) gLink->bufFrame)->magic == NDXMAGIC);
	
	INDEXDESC* np = & (((INDEXDESCPAGE*) gLink->bufFrame)->desc);
	
	/*
	 *	Check parameters
	 */
	if (bound1 && bound1->length > np->maxKeyLen)  {
		SM_ERROR(TYPE_USER, esmKEYTOOLONG);
		bf_UnfixPage(gLink, BF_LRU, FALSE);
		return esmFAILURE;
		}
	if (bound2 && bound2->length > np->maxKeyLen)  {
		SM_ERROR(TYPE_USER, esmKEYTOOLONG);
		bf_UnfixPage(gLink, BF_LRU, FALSE);
		return esmFAILURE;
		}

	SM_ASSERT(LEVEL_3, np->maxKeyLen <= SM_MAXKEYLEN);
	
	if ((cond1&SM_G && !(cond2&SM_L)) || (cond2&SM_G && !(cond1&SM_L))) {
		SM_ERROR(TYPE_USER, esmCONFLICTCOND);
		bf_UnfixPage(gLink, BF_LRU, FALSE);
		return esmFAILURE;
		}
	
	/*
	 *	Initialize cursor
	 */
	cursor->ndx = *ndx;
	cursor->groupIndex = groupIndex;
	cursor->eof = FALSE;
	cursor->cond2 = cond2;
	if (! bound2) 
		cursor->b2len = 0;
	else {
		BCOPY(bound2->valuePtr, cursor->b2, bound2->length);
		cursor->b2len = bound2->length;
		}
	BT_CURSOR* btCursor = & (cursor->opaque.btCursor);
	
	
	/*
	 *	Initialize the fetch
	 */
	void* b1Ptr = 0;
	int   b1Len = 0;
	
	if (bound1)  {
		b1Ptr = bound1->valuePtr;
		b1Len = bound1->length;
		}

	switch (np->ndxType)  {
		case SM_BTREENDX:
			if (bt_FetchInit(np->rootPid, bufGroup, b1Ptr, b1Len, 
									np->maxKeyLen, btCursor, cond1, 
									SMCOMPFUNC[np->dataType])) {
				bf_UnfixPage(gLink, BF_LRU, FALSE);
				return esmFAILURE;
				}
			break;

		case SM_HASHNDX:
			if (cond1 != SM_EQ || cond2 != SM_EQ || 
					bound1->length != bound2->length) {
				SM_ERROR(TYPE_USER, esmCONFLICTCOND);
				bf_UnfixPage(gLink, BF_LRU, FALSE);
				return esmFAILURE;
				}
			if (bound1->valuePtr != bound2->valuePtr &&
					bcmp((char*)bound2->valuePtr, (char*)bound1->valuePtr, bound2->length)) {
				SM_ERROR(TYPE_USER, esmCONFLICTCOND);
				bf_UnfixPage(gLink, BF_LRU, FALSE);
				return esmFAILURE;
				}

			if (lh_FetchInit(np->rootPid, bufGroup, b1Ptr, b1Len,
								np->maxKeyLen, btCursor, cond1,
								SMCOMPFUNC[np->dataType])) {
				bf_UnfixPage(gLink, BF_LRU, FALSE);
				return esmFAILURE;
				}
			break;

		default:
			SM_ERROR(TYPE_USER, esmWRONGNDXTYPE);
		}

	bf_UnfixPage(gLink, BF_LRU, FALSE);
	
	if ( ! btCursor->valid)  {
		cursor->eof = TRUE;
		}

	return esmNOERROR;
}

