/*
 *   $RCSfile: sm_Initialize.c,v $  
 *   $Revision: 2.35 $  
 *   $Date: 1993/03/08 21:00:09 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "sysdefs.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "sm_macro.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "sm_params.h"
#include "btree.h"
#include "volume.h"
#include "lock.h"
#include "msgdefs.h"
#include "lsn.h"
#include "loginfo.h"
#include "file.h"
#include "serverinfo.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "fi_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#include "trans_funcs.h"
#include "lg_extfuncs.h"
#include "msg_funcs.h"
#include "trace_funcs.h"
#ifdef JUNK
#include "sequentialfile.h"
#include "sequential_funcs.h"
#endif JUNK
#include "clog.h"
#include "clog_extfuncs.h"
#include "io_globals.h"
#include "sm_globals.h"


 int
sm_Initialize ()
{
	static PFI initfuncs[] = {
		SM_InitClientOptions,
		initTrace,
		SM_CheckRequiredOptions,
		SM_Init,
		io_Init,
		bf_Init,
		fi_InitFileInfoTable,
		initTrans,
#ifdef JUNK
		initSequential,
#endif JUNK
		lg_Init,
		initServerInfo,
		initMask,
		initializeLinks,
		clog_Initialize,
		(PFI) 0
	}; 
	PFI *func;

	/* set the error stream if not already set by the application */
	if (sm_ErrorStream == NULL) sm_ErrorStream = stderr;

	/*
	 *	Report that the LGTESTING is defined in object.h and therefore
	 *	large object btree nodes are too small
	 */
#ifdef LGTESTING
	fprintf(stderr, "WARNING: The SM has been compiled to use\n");
	fprintf(stderr, "         low fanout large obj btree nodes.\n");
#endif

	/*
	 *	Enforce the requirement that the minimum size of a small
	 *	object must be enough to hold a LARGEOBJHDR.  See object.h
	 *	for more details.
	 *
	 *	Also, object headers must have an aligned size.
	 */
	SM_ASSERT( LEVEL_1, sizeof(OBJECT) >= sizeof(LARGEOBJHDR) );
	SM_ASSERT( LEVEL_1, sizeof(OBJHDR) % ALIGNON == 0 );

	(void)BZERO(&MsgStats, sizeof(MsgStats));

	for(func = initfuncs; *func != (PFI) NULL; func++) {
		if( (*(*func))() != esmNOERROR ) {
			(void) sm_ShutDown();
			return esmFAILURE;
		}
	}
	return esmNOERROR;
}

 int
sm_ShutDown ()
{
	TRACE(TR_SM, TR_LEVEL_1);

    switch (TransRec.clientTransState) {

    case T_ACTIVE:
    case T_PREPARED:

        SM_ERROR(TYPE_WARNING, esmTRANSINPROGRESS);

    case T_ABORT:
    case T_INACTIVE:

		break;

    default:
        SM_ERROR(TYPE_FATAL, esmINTERNAL);
        return esmFAILURE;
    }

	shutServerInfo();
	trans_Shutdown();

	bf_Shutdown();

	/* btree_Shutdown(); */
	clog_Shutdown();
	io_Shutdown();
#ifdef JUNK
	shutdownSequential();
#endif JUNK
	fi_FreeFileInfoTable();
	links_Shutdown();

	SM_Shutdown();
	SM_FREE_PTR(VersionGraphSpace);

	sm_ErrorStream = NULL;
	return esmNOERROR;
}
