/*
 *   $RCSfile: servermain.c,v $  
 *   $Revision: 2.33 $  
 *   $Date: 1993/04/14 18:36:12 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "sysdefs.h"
#include "ess.h"
#include <fcntl.h>
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "object.h"
#include "lock.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "bitvec.h"
#include "link.h"
#include "msgvector.h"
#include "bf.h"
#include "init_funcs.h"
#include "server_options.h"
#include "server_main_funcs.h"
#include "sharedmem_globals.h"
#include "volume.h"
#include "lsn.h"
#include "openlog.h"
#include "msg_globals.h"
#include "bf_globals.h"
#include "io_globals.h"
#include "log_globals.h"
#include "msg_globals.h"
#include "ips_support.h"

main(int argc, char** argv)
{

/*
 *  Here we check the sizes of the basic datatypes used in
 *  slotted pages to make sure they are all multiple of the
 *  object alignment macro ALIGNON (usually 8).  Hopefully,
 *  the compiler's constant folding will report problems
 *  immedidately.
 */
#ifdef DEBUG
int debugunused1 = 1/ ( (sizeof(OBJHDR) % ALIGNON) == 0 ? 1: 0);
int debugunused2 = 1/ ( (sizeof(LARGEOBJROOT) % ALIGNON) == 0 ? 1: 0);
int debugunused3 = 1/ ( (sizeof(LARGEOBJHDR) % ALIGNON) == 0 ? 1: 0);
int debugunused4 = 1/ ( (sizeof(LGDATAHDR) % ALIGNON) == 0 ? 1: 0);
#endif
SM_ASSERT(LEVEL_1, (sizeof(OBJHDR) % ALIGNON) == 0);
SM_ASSERT(LEVEL_1, (sizeof(LARGEOBJROOT) % ALIGNON) == 0);
SM_ASSERT(LEVEL_1, (sizeof(LARGEOBJHDR) % ALIGNON) == 0);
SM_ASSERT(LEVEL_1, (sizeof(LGDATAHDR) % ALIGNON) == 0);

#ifdef IPS_SUPPORT
	IPSuareaReadable = 0;
#endif IPS_SUPPORT

	/*
	 *	By default, use the the user id as the shared memory key.
	 */
	ShmKey = getuid();

	/*
	 *	Default error file is stderr
	 *      Make stdout line buffered when
	 *      it is a file (vs. default, only if a terminal).
	 */
	sm_ErrorStream = stderr;
	setlinebuf(stdout);
	/* 
	 * don't want disk processes to get SIGTTIN
	 * so set stdin to be close-on-exec, and
	 * We ignore the TTIN, so that we just get EOF
	 * (see catchSignals.c)
	 */
	if (fcntl(fileno(stdin), F_SETFD, NULL) < 0)    {
		if(errno == EBADF) {
			/* oh well, it's closed by the calling shell */
		} else {
			SM_ERROR(TYPE_FATAL, errno);
		}
	}

	/*
	 * map error codes to string messages
	 */
	initErrorCodes();

	initializeVolNameTable();
	/*
	 *	Evaluate command line options and process all configuration
	 *	files.
	 */
	processCommandLine(argc, argv);

	/*
	 *	Begin initialization
	 */
	initializeServer();

	/* if recover included PREPARED tx's, we still have these volumes
	 * mounted
	 */
}
