/*
 *   $RCSfile: admin_ShutServer.c,v $  
 *   $Revision: 1.45 $  
 *   $Date: 1993/04/21 17:32:48 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "latch.h"
#include "bf.h"
#include "volume.h"
#include "trans.h"
#include "openlog.h"
#include "logrecs.h"
#include "bitvec.h"
#include "adminmsg.h"
#include "disk.h"
#include "threadstate.h"
#include "trans_intfuncs.h"
#include "trans_extfuncs.h"
#include "lm_extfuncs.h"
#include "log_extfuncs.h"
#include "log_intfuncs.h"
#include "msg_funcs.h"
#include "admin_funcs.h"
#include "io_extfuncs.h"
#include "io_intfuncs.h"
#include "disk_funcs.h"
#include "bf_extfuncs.h"
#include "server_util_funcs.h"
#include "util_funcs.h"
#include "queue_consist.h"
#include "distr.h"
#include "thread_funcs.h"
#include "log_globals.h"
#include "io_globals.h"
#include "disk_globals.h"
#include "msg_globals.h"
#include "thread_globals.h"
#include "trans_globals.h"

static BOOL 
listvol(
	VOLREC  *volRec,
	int		*num_open,
	int		*total
)
{
	fprintf(stdout, 
		"VOLID=%d name=%s ",
		volRec->volid, volRec->volNameRec->volName.name);

	if(volRec->volflags & VOL_OPEN) {
		fprintf(stdout, "OPEN ");
		(*num_open)++;
	}
	if(volRec->volflags & VOL_TEMP) {
		fprintf(stdout, "TEMP ");
	}
	if(volRec->volflags & VOL_LOG) {
		fprintf(stdout, "LOG ");
	}
	if(volRec->volflags & VOL_RAWDEV) {
		fprintf(stdout, "RAWDEV ");
	}
	if(volRec->volflags & VOL_INIT) {
		fprintf(stdout, "INIT ");
	}
	if(volRec->volflags & VOL_FORMAT) {
		fprintf(stdout, "FORMAT ");
	}
	if(volRec->volflags & VOL_QUIESCE) {
		fprintf(stdout, "QUIESCE ");
	}
	if(volRec->volflags & VOL_INVALID) {
		fprintf(stdout, "INVALID ");
	}
	if(volRec->volflags & VOL_DISPLAY) {
		fprintf(stdout, "DISPLAY ");
	}
	/* any other bits? */
	if(volRec->volflags & ~(
		VOL_LOG | 
		VOL_TEMP | 
		VOL_DISPLAY | 
		VOL_RAWDEV | 
		VOL_OPEN | 
		VOL_INIT | 
		VOL_QUIESCE | 
		VOL_FORMAT | 
		VOL_INVALID 
	)) {
		fprintf(stdout, "(0x%x)", volRec->volflags);
	}
	fprintf(stdout, "\n");

	(*total)++;
		
	return FALSE;
}

void
admin_ListMount ()
{
	int total = 0;
	int num_open = 0;
	int left;

	ForEachVolume(3, listvol, &num_open, &total, 0 );

	/* NumVolumes already has MaxAddVolumes added in (at init time) */
	left = NumVolumes - total;

	fprintf(stdout, 
		"*** %d volume%s mounted and open; %d mount slot%s in use; %d mount slot%s free\n",
		num_open, (num_open == 1)?"":"s",
		total, (total == 1)?"":"s",
		left, (left == 1)?"":"s");
}

static BOOL 
listvolname(
	VOLNAMEREC  *volNameRec,
	int		*total
)
{
	fprintf(stdout, "%s:%d:%d:%d:%d",
		volNameRec->volName.name,
		volNameRec->volid,
		volNameRec->fmtinfo->numCylinders,
		volNameRec->fmtinfo->tracksPerCylinder,
		volNameRec->fmtinfo->pagesPerTrack
		);

	if(volNameRec->fmtinfo->numLogFileBlocks != 0) {
		fprintf(stdout, ":%d\n", 
			volNameRec->fmtinfo->logFileBlock2size >> LOG2_OF_1K);
		SM_ASSERT(LEVEL_1, (volNameRec->fmtinfo->numLogFileBlocks == -1));
		SM_ASSERT(LEVEL_1, (volNameRec->fmtinfo->volflags & VOL_LOG));
		SM_ASSERT(LEVEL_1, (volNameRec->fmtinfo->volflags & VOL_TEMP)==0);
	} else {
		SM_ASSERT(LEVEL_1, (volNameRec->fmtinfo->volflags & VOL_LOG)==0 );
		/* could be temp or data */
		fprintf(stdout, "\n");
	}

	(*total)++;
		
	return FALSE;
}

void
admin_ListVolNames ()
{
	int total = 0;

	ForEachVolumeNameRec(2, listvolname, &total, 0, 0 );

	fprintf(stdout, 
		"*** %d volume%s \n",
		total, (total == 1)?"":"s");
}


static BOOL 
shutnonlogvol(
	VOLREC 	*volRec,
	BOOL	*recoveryNeeded,
	FLAGS	*flags
)
{
	if( (volRec == OpenLog.volRec) ) 
		return FALSE;
	if( !(volRec->volflags & VOL_OPEN) ) 
		return FALSE;

	/*
	 * 	See if any active transactions on this volume	
	 */
	if (LIST_NOT_EMPTY(&(volRec->activeTransList)) || 
							!(*flags & SHUT_CLEAN_VOLUMES)) {
		fprintf(stdout, 
			"Warning: volume %d:%s requires recovery processing.\n", 
			volRec->header->volid, volRec->volNameRec->volName.name);

		if (volRec->header->volid == OpenLog.volid) {
			fprintf(stdout, "\tNote: the above volume contains the log.\n");
		}
		*recoveryNeeded = TRUE;

	} else if ( (*flags & SHUT_CLEAN_VOLUMES) && (volRec->header->volid != OpenLog.volid) ) {
	
		/*
		 *	Unhook any client link and then dismount the
		 *	volume, and log the dismount (assuming the
		 *	volume is not the log volume).
		 */
		initializeList(&(volRec->openList));
		if (io_DismountVolume(volRec, TRUE) != esmNOERROR) {
			fprintf(stdout, 
				"Warning: volume %d:%s requires recovery processing.\n",
				volRec->header->volid, volRec->volNameRec->volName.name);
		}
	}
	return FALSE;
}

void
admin_ShutServer (
	FLAGS	flags 
)
{
	TRANSREC	*transRec;
	BOOL		recoveryNeeded;
	char		*pointer;

	TRPRINT(TR_ADMIN, TR_LEVEL_1, 
			("shutting down server with flags 0x%x", flags));

	if(flags == 0) {
		flags = SHUT_ABNORMAL_EXIT;
	}
	/*
	 *	If this is the formatvol program, just exit.
	 */
	if (formatvolRunning) {
		exit(Active->errno);
	}

	/*
	 *	make sure parameters make sense
	 */
	if(flags & SHUT_ABNORMAL_EXIT) {
		SM_ASSERT(LEVEL_1,(flags & ~(SHUT_ABNORMAL_EXIT | SHUT_DUMP_CORE))==0);
	}
	if ( (flags & SHUT_COMMIT_TRANS) && (flags & SHUT_ABORT_TRANS) ) {
		SM_ERROR(TYPE_USER, esmBADPARAMS);
	}

	/*
	 *	Mark that transaction processing shutdown is in effect.
	 */
	if (TransShutdownState != T_ACTIVE) {
		fprintf(stdout, 
			"Failure in server shutdown (or multiple shutdown requests).\n");
		fprintf(stdout, 
			"Warning: immediate shutdown --");
		fprintf(stdout, 
			"recovery required on all mounted volumes.\n");
		goto do_exit;
	}
	TransShutdownState = T_INACTIVE;

	/*
	 *	See if we are supposed to commit all currently running 
	 *	transactions.
	 */
	if (flags & SHUT_COMMIT_TRANS) {

		/*
		 *	Mark that transaction processing shutdown is in effect.
		 */
		TransShutdownState = T_COMMIT;

		/*
		 *	Wait for all transactions to end
		 */
		if (LIST_NOT_EMPTY(&ActiveTransList)) {
			fprintf(stdout, 
				"Transactions are in progress. ");
			fprintf(stdout,
				"Waiting for them to commit...\n");
			if (waitList(&TransShutdownWaitList, 
				THREAD_TRANS_SHUTDOWN_WAIT) != esmNOERROR) {
				SM_ERROR(TYPE_STOP, esmINTERNAL);	
			}
		}
	}

	/*
	 *	See if we are supposed to abort all currently running 
	 *	transactions.
	 */
	if (flags & SHUT_ABORT_TRANS) {

		/*
		 *	Mark that transaction processing shutdown is in effect.
		 */
		TransShutdownState = T_ABORT;

		/*
		 *	Abort all currently running transactions
		 */
		transRec = (TRANSREC *) FIRST_LIST_ELEMENT(&ActiveTransList);
		if(transRec != NULL ) {
			fprintf(stdout, 
				"Transactions are in progress. ");
			fprintf(stdout,
				"Aborting them ...\n");
		}
		while (transRec != NULL)    {
			CHECK_TRANSREC_MAGIC(transRec);
			internalAbort(transRec, esmSERVERDIED); /* ignores those PREPARED, END, COMMIT */
			transRec = (TRANSREC *) 
					NEXT_LIST_ELEMENT( &(transRec->activeTransList));
		}

		/*
		 *	Wait for all transactions to end
		 */
		for(;;) {
			BOOL	done;

			/* go through the list: 
			 * "done" indicates whether the list
			 * contains any non-2pc transactions
			 */
			for (done=TRUE, transRec = (TRANSREC *) FIRST_LIST_ELEMENT(&ActiveTransList);
					transRec != NULL;
					transRec = (TRANSREC *) NEXT_LIST_ELEMENT( &(transRec->activeTransList))
				) {

				if( ! TRANS_ALIVE_AFTER_RECOVERY(transRec) ) {
					done = FALSE;
				}
			}
			if(done) 
				break;
			Sleep(T_SHUTDOWNTIMEOUT);
		}
	}

	/*
	 *	See if a checkpoint was reqeusted
	 */
	if (flags & (SHUT_TAKE_CHECKPOINT | SHUT_CLEAN_VOLUMES)) {

		/*
		 *	Wait for any current checkpoint to complete
		 */
		if (OpenLog.flags & LOG_CHECKPOINT_IN_PROGRESS) {
			fprintf(stdout, 
				"Checkpoint is in progress. ");
			fprintf(stdout,
				"Waiting for it to complete ...\n");
			if (waitList( &(OpenLog.checkPointList), THREAD_CHECKPOINT_WAIT )){
				SM_ERROR(TYPE_STOP, esmINTERNAL);	
			}
		}
		/* start a checkpoint and wait for it to complete */
		fprintf(stdout, 
			"Starting a checkpoint. ");
		fprintf(stdout,
			"Waiting for it to complete ...\n");
		activateCheckpoint(TRUE, flags & SHUT_CLEAN_VOLUMES, TRUE);
		/* There are no dirty pages left; a checkpoint won't be
		 * doing anything to a volume after we close the volume,
		 * so we don't have to worry about losing the CPU while
		 * closing volumes.
		 */
	} 

	/*
	 *	Report on any messed up volumes.  A volume is messed up if
	 *	dirty pages were left in the buffer pool for it, or if a
	 *	transaction was active which required it.
	 */
	recoveryNeeded = FALSE;

	ForEachVolume(3, shutnonlogvol, &recoveryNeeded, &flags, 0 );

	if (!recoveryNeeded) {
		fprintf(stdout, 
			"Clean shutdown: no recovery required on any volumes.\n");
	}

	/*
	 *	Shutdown the log volume
	 */
	if (OpenLog.volRec) {
		if (flags & SHUT_CLEAN_VOLUMES) {

			/*
			 *	Remove the log volume from its permanent link
			 */
			unlinkVolume( &LogLink, OpenLog.volRec );

			/*
			 *  unfix the checkpoint info
			 */
			bf_UnfixPage(OpenLog.controlBuffer, BF_DEFAULT, FALSE);

			if (bf_DismountVolume(OpenLog.volRec, FALSE) != esmNOERROR) {
				fprintf(stdout, "Warning: volume %d:%s requires recovery processing.\n", 
					OpenLog.volRec->header->volid, 
					OpenLog.volRec->volNameRec->volName.name);
			}
		} else if( !(flags & SHUT_ABNORMAL_EXIT) ) {
			/* have to remove it from table first */
			IO_RemoveVolRec(OpenLog.volRec);
			closeDisk(OpenLog.volRec);
		}
	}

do_exit: 
	killDiskProcs( );

	/*
	 * free shared-memory, IPC resources 
	 */
	removeIpcs();

	/*
	 * In case it hasn't been done yet, close all links to clients
	 */
	{
		/* index into Links[] is fd */
		int i;
		for(i=3;  i<NumLinks; i++) {
			if( Links[i].flags == LINK_FREE ) {
				/* not used */
				continue;
			}
			if( Links[i].linkClass == 0 ) {
				/* not used */
				continue;
			}
			TRPRINT(TR_ADMIN, TR_LEVEL_1, 
				("closing fd %d, class %d flags 0x%x", 
					i, Links[i].linkClass,
					Links[i].flags));
			/*
			 * no need to clearLink(), because only
			 * the parent uses the SelectMask, DiskMask, etc
			 */
			if (Links[i].linkClass == CL_CLIENT) {
				if( shutdown(i,2) < 0) {
					SM_ERROR(TYPE_SYS, errno);
				}
			} else {
				if( close(i) < 0) {
					SM_ERROR(TYPE_SYS, errno);
				}
			}
		}
	}

#ifdef DEBUG
	fprintf(stderr, "Largest stack used: %d words.\n", LargestStack);
#endif DEBUG

	/*
	 *	We are finally going to exit.  First see if we should
	 *	dump core.
	 */
	if (flags & SHUT_DUMP_CORE) {
		fprintf(stdout, "The server is dumping core for your debugging convenience.\n");
#ifdef notdef
		unCatchAbortSignal(SIGILL);
		abort();
#else notdef
		unCatchAbortSignal(SIGSEGV);
		pointer = NULL;
		*pointer = 0;
#endif notdef
	} else {
		exit(recoveryNeeded);
	}
	/*NOTREACHED*/
}
