/*
 *   $RCSfile: BF_InitializeSharedMemory.c,v $  
 *   $Revision: 1.14 $  
 *   $Date: 1993/04/13 19:56:35 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "io.h"
#include "tid.h"
#include "lock.h"
#include "latch.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "bf.h"
#include "bf_intfuncs.h"
#include "bf_globals.h"
#include "sharedmem_globals.h"

 void 
BF_InitializeSharedMemory(

	int		size 
)
{

	int	try_count;	/* for multiple tries to shmget	*/
#ifdef vax
	char    *shmSpace;
#endif vax

	TRPRINT(TR_INIT, TR_LEVEL_1, ("allocating:%d bytes", size));

	/*
	 *	create the shared memory segment, if it already exists,
	 *	but is the wrong size, delete it and try_count again
	 */
	try_count = 0;
	while (try_count < 2) {

#ifdef mips
#define SHM_R   0400    /* read permission */
#define SHM_W   0200    /* write permission */
#define SM_R    0400
#define SM_W    0200

		ShmId = shmget(ShmKey, size, IPC_CREAT|SM_R|SM_W);
#elif defined(sparc) || defined(_AIX) || defined(hpux)
/* some include files don't define our permissions for us... */
#ifndef SHM_W
#define SHM_W 0200
#endif
#ifndef SHM_R
#define SHM_R 0400
#endif
		ShmId = shmget(ShmKey, size, IPC_CREAT|SHM_R|SHM_W);
#elif defined(vax)
		ShmId = shmget(ShmKey, size, IPC_CREAT|SM_R|SM_W);
#elif !(defined(mips) || defined(sparc) || defined(vax) || defined(_AIX) || defined(hpux))
	not supported
#else
	/* this else should have the not supported code, but the
	   ultrix cpp is broken */
#endif
		if (ShmId < 0)	{

			if (try_count == 1) {

				/*
				 *	No success
				 */
				SM_ERROR(TYPE_FATAL, errno);
			}

			/*
			 *	Try again to get the segment, without creating it
			 */
#ifdef mips
			ShmId = shmget(ShmKey, 0, SM_R|SM_W);
#elif defined(sparc) || defined(_AIX) || defined(hpux)
			ShmId = shmget(ShmKey, 0, SHM_R|SHM_W);
#elif defined(vax)
			ShmId = shmget(ShmKey, 0, SM_R|SM_W);
#elif !(defined(mips) || defined(sparc) || defined(vax) || defined(_AIX) || defined(hpux))
	not supported
#else
	/* this else should have the not supported code, but the
	   ultrix cpp is broken */
#endif
			if (ShmId < 0)	{
				SM_ERROR(TYPE_FATAL, errno);
			}

			/*
			 *	The segment exists, but may not be large enough, so
			 *	delete the segment 
			 */
			if (shmctl(ShmId, IPC_RMID, NULL)) {
				SM_ERROR(TYPE_FATAL, errno);
			}

		} else {
			
			/*
			 *	Success, so leave the try_count loop
			 */
			break;
		}

		/*
		 *	try_count again
		 */
		try_count++;
	}
	TRPRINT(TR_INIT, TR_LEVEL_2, ("shared memory id:%d", ShmId));

#if defined(mips) || defined (sparc) || defined(_AIX) || defined(hpux)

	/*
	 *	attach the shared memory segment
	 */
	if ((ShmAddress = shmat(ShmId, 0, 0)) == (char *) -1)	{

		SM_ERROR(TYPE_FATAL, errno);
	}

#elif defined(vax)

    /*
     *  Attach the shared memory segment and avoid bug in shmat.
     *  Apparently shmat interferes with malloc on the vax so
     *  here we allocate the space for the shared memory ourselves
     *  and tell the os to put the shared memory there.
     *  Actually the above idea doesn't work so instead we'll put it
     *  10Meg out in the address space.
     */

    /*
     * make it plenty big
     * shmSpace = (char*)(((long)(malloc(size*2) + 4096)) & 0xffffe000);
     */

    /*
     * put it far out
     */
    shmSpace = (char*)(10 * 1024 * 1024); /* 10 meg */
    if ((ShmAddress = shmat(ShmId, shmSpace, SHM_RND)) == (char *) -1) {

        SM_ERROR(TYPE_FATAL, errno);
    }
#elif !(defined(mips) || defined(sparc) || defined(vax) || defined(_AIX) || defined(hpux))
	not supported
#else
	/* this else should have the not supported code, but the
	   ultrix cpp is broken */
#endif

	/*
	 *	Record the shared memory segment size
	 */
	ShmSize = size;

	TRPRINT(TR_INIT, TR_LEVEL_2, ("shared memory attached at:%x", ShmAddress));
}
