/*
 *   $RCSfile: cl_Rusage.c,v $  
 *   $Revision: 1.10 $  
 *   $Date: 1993/04/14 19:01:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "sysdefs.h"
#include <sys/resource.h>
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "io.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#include "msgdefs.h"
#include "thread.h"
#include "thread_globals.h"
#include "semaphore.h"
#include "bitvec.h"
#include "latch.h"
#include "link.h"
#include "cl_funcs.h"
#include "msg_funcs.h"

static struct rusage since;

#define diffit(new,e,b,x) new.x = e.x - b.x

#define difftime(new,e,b,x) {\
	register int z = ((((e.x.tv_sec)*1000000) + e.x.tv_usec) - \
			(((b.x.tv_sec)*1000000) + b.x.tv_usec)); \
	new.x.tv_sec = z / 1000000;\
	new.x.tv_usec = z % 1000000;\
}


#define DIFFRUSAGE(new,e,b)\
	difftime(new,e,b,ru_utime);\
	difftime(new,e,b,ru_stime);\
	diffit(new,e,b,ru_minflt);\
	diffit(new,e,b,ru_majflt);\
	diffit(new,e,b,ru_nswap);\
	diffit(new,e,b,ru_inblock);\
	diffit(new,e,b,ru_oublock);\
	diffit(new,e,b,ru_msgsnd);\
	diffit(new,e,b,ru_msgrcv);\
	diffit(new,e,b,ru_nsignals);\
	diffit(new,e,b,ru_nvcsw)

 void
cl_ResetRusage (
	MESSAGE		*unused1,
	LINK		*unused2 
)
{
	if(getrusage(RUSAGE_SELF, &since) < 0) {
		reply(esmFAILURE, errno, 1, TRUE);
	}
	reply(esmNOERROR, esmNOERROR, 1, TRUE);
}

 void
cl_GetRusage (
	MESSAGE		*unused1,
	LINK		*unused2 
)
{
	struct rusage new_rusage;

	errno = 0;

	if(getrusage(RUSAGE_SELF, &new_rusage) < 0) {
		reply(esmFAILURE, errno, 1, TRUE);
	}
	DIFFRUSAGE(new_rusage,new_rusage,since);

	Active->iovec[1].iov_base = (char *)&new_rusage;
	Active->iovec[1].iov_len = (int) sizeof(new_rusage);

	reply(esmNOERROR, esmNOERROR, 2, TRUE); 
}
