/*
 *   $RCSfile: forkDiskProc.c,v $  
 *   $Revision: 2.34 $  
 *   $Date: 1993/04/21 17:38:27 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "sysdefs.h"
#include <fcntl.h>
#if defined(sparc)
#include <vfork.h>
#endif
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "latch.h"
#include "bitvec.h"
#include "disk.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "bf.h"
#include "pool.h"
#include "volume.h"
#include "disk_funcs.h"
#include "msg_funcs.h"
#include "msgvector.h"
#include "msg_globals.h"
#include "disk_globals.h"
#include "io_globals.h"
#include "sharedmem_globals.h"
#include "queue_globals.h"
#include "queue_consist.h"
#include "admin_funcs.h"


extern int			ShmId;

struct shmid_ds		shmstruct;
static char			SocketName[20];
static char			ShmIdName[20];

static char			SemIdName[20];
static char			SemNumStr[20];
static char			MsgBufsNum[20];
static char			DiskQAddr[20];
static char			RepliesAddr[20];
static char			TraceFlagsName[20];

#if (defined (c_plusplus) & !defined(__cplusplus)) || defined(sparc)

#	define smem		shmid_ds    
#	define long		key_t       
#	define sm_perm	shm_perm    
#	define sm_size	shm_segsz   
#	define sm_cpid	shm_cpid    
#	define sm_lpid	shm_lpid    
#	define sm_count	shm_nattch  
#	define sm_atime	shm_atime   
#	define sm_dtime	shm_dtime   
#	define sm_ctime	shm_ctime   

#endif 

int
forkDiskProc ( 
	VOLREC *volRec
)

{
	register LINK	*link;
	register int	child;
	int				diskSocket[2];

	SM_ASSERT(LEVEL_1, (FastPathOnly == FALSE));

	TRPRINT(TR_DISKRW, TR_LEVEL_1, ("forking disk process"));

	/*
	 *	create the socket pair
	 */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, diskSocket))	{

		SM_ERROR(TYPE_SYS, errno);
		return esmFAILURE;
	}
	TRPRINT(TR_DISKRW|TR_MSG, TR_LEVEL_2, ("parent socket:%d child socket:%d",
			diskSocket[1], diskSocket[0]));

	/*
	 *	setup the link structure
	 *  Now, the child will overwrite the link ptr, but we need
	 *  it set up in the parent, whether or not we do the fork,
	 *  so we do it here.
	 */
	if (diskSocket[1] >= NumLinks) {
		SM_ERROR(TYPE_CRASH, esmINTERNAL);
	}
	link = &(Links[diskSocket[1]]);	
	link->linkClass = CL_DISK;
	setLink(link);
	volRec->volLink =  link;

	/*
	 *	fork off the child process
	 *  TODO: probably ought to stat DiskProcName first!
	 */

	SM_ASSERT(LEVEL_3, DiskProcName != NULL);

#define VFORK

#ifdef VFORK
#define sysfork vfork
#else 
#define sysfork fork
#endif
	if (child = sysfork())	{

		/*
		 *	we are in the PARENT
		 *
		 *	check to see if the fork succeeded
		 */
		if (child == -1)	{

			SM_ERROR(TYPE_SYS, errno);
			return esmFAILURE;
		}


		/*
		 *	close child end
		 */
		close(diskSocket[0]);

		/*
		 *	set the socket to close on exec for disk processes
		 *  so that when we fork for *another* disk process, this
		 *  socket won't be open in that disk process.
		 */
		if (fcntl(diskSocket[1], F_SETFD, NULL) < 0)	{

			SM_ERROR(TYPE_FATAL, errno);
		}

		TRPRINT(TR_DISKRW, TR_LEVEL_1, ("Parent %d forks %s %d for %s %d\n",
			getpid(), DiskProcName, child, "disk#", volRec->queues->semNum));

		volRec->diskrwPid = child;
		return esmNOERROR; /* a-ok */
	} else {
		/*
		 *	we are in the CHILD
		 */

		/* Old : just close the other end of the pipe: 
		 *	close(diskSocket[1]);
		 * New : close everything we will not use:
		 */
		{
			/* index into Links[] is fd */
			register int i;
			for(i=3;  i<NumLinks; i++) {
				if(i != diskSocket[0]) {
					if( Links[i].flags == LINK_FREE ) {
						/* not used */
						continue;
					}
					if( Links[i].linkClass == 0 ) {
						/* not used */
						continue;
					}
					TRPRINT(TR_DISKRW, TR_LEVEL_1, 
						("closing fd %d, class %d flags 0x%x", 
							i, Links[i].linkClass,
							Links[i].flags));
					/*
					* can't clearLink(), because of vfork, besides,
					* only the parent uses the SelectMask, DiskMask, etc
					*/
					if( close(i) < 0) {
				 		/* can't be TYPE_FATAL| 
						TYPE_QUIET | TYPE_STOP | TYPE_CRASH*/
						 SM_ERROR(TYPE_SYS, errno); 
						/* sets Active->errno in parent */
						_exit(errno);
					}
				} 
			}
		}


		/*
		 *	get shared memory information
		 */
		if (shmctl(ShmId, IPC_STAT, &shmstruct) < 0)	{
			 /* can't be TYPE_FATAL| TYPE_QUIET | TYPE_STOP | TYPE_CRASH*/
			 SM_ERROR(TYPE_SYS, errno); /* sets Active->errno in parent */
			_exit(errno);
		}
		TRPRINT(TR_SHM, TR_LEVEL_2, ("shared memory Id:%x", ShmId));
#if defined(_AIX) || defined(hpux)
		TRPRINT(TR_SHM, TR_LEVEL_2, ("shared memory size:%d", 
				shmstruct.shm_segsz));
		TRPRINT(TR_SHM, TR_LEVEL_2, ("shared memory reference count:%d",
				shmstruct.shm_nattch));
#else 
		TRPRINT(TR_SHM, TR_LEVEL_2, ("shared memory size:%d", 
				shmstruct.sm_size));
		TRPRINT(TR_SHM, TR_LEVEL_2, ("shared memory reference count:%d",
				shmstruct.sm_count));
#endif

		/*
		 *	format the appropriate strings
		 */
		sprintf(TraceFlagsName, "%x", TraceFlags);
		sprintf(ShmIdName, "%x", ShmId);
		sprintf(SocketName, "%x", diskSocket[0]);
		sprintf(SemIdName, "%x", SemId);
		sprintf(SemNumStr, "%x", volRec->queues->semNum);
		sprintf(DiskQAddr, "%x", ShmAddressToOffset(volRec->queues) );
		sprintf(RepliesAddr, "%x", ShmAddressToOffset(Replies) );
		sprintf(MsgBufsNum, "%x", NumMsgBufs);

		TRPRINT(TR_DISKRW, TR_LEVEL_2, 
		("execl(%s %s %s %s %s %s %s %s %s %s 0)",
		/*    	0  1  2  3  4  5  6  7  8  9   	*/ 
			/* 0 */ DiskProcName, 
			/* 1 */ TraceFlagsName, 
			/* 2 */ ShmIdName, 
			/* 3 */ SocketName, 
			/* 4 */ volRec->volNameRec->volName.name,
			/* 5 */ SemIdName,  
			/* 6 */ SemNumStr,  
			/* 7 */ DiskQAddr,
			/* 8 */ RepliesAddr,
			/* 9 */ MsgBufsNum));

		/*
		 *	exec the routine
		 */
		if (execl(DiskProcName, 
			/* 0 */ DiskProcName, 
			/* 1 */ TraceFlagsName, 
			/* 2 */ ShmIdName, 
			/* 3 */ SocketName, 
			/* 4 */ volRec->volNameRec->volName.name,
			/* 5 */ SemIdName,  
			/* 6 */ SemNumStr,  
			/* 7 */ DiskQAddr,
			/* 8 */ RepliesAddr,
			/* 9 */ MsgBufsNum,
				0) < 0)	{

			fprintf(sm_ErrorStream, 
				"SERVER_ERROR: Could not execute the disk process: %s\n", 
				DiskProcName);
			fprintf(sm_ErrorStream, 
				"\tPlease make sure %s is accessible.\n", DiskProcName);

			if (kill(getppid(), SIGINT) < 0) {
				fprintf(sm_ErrorStream, 
					"Could not kill parent (server) process: %d.\n", 
					getppid());
			}
			removeIpcs();
		}
	}
	return esmNOERROR; /* to keep g++ quiet */
}
