#ifdef MEDIA_RECOVERY
/*
 *   $RCSfile: connect_routines.c,v $  
 *   $Revision: 1.14 $  
 *   $Date: 1993/02/04 20:30:06 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
/* All the include files */

#define MAXHOSTNAMELEN 255

#include "sysdefs.h"
#include "ess.h"
#include "io.h"
#include "com_packet.h"
#include "mr_intfuncs.h"

extern int errno;


/* ========================================================================= */

/* This function is called to set up a TCP socket which can later be used for
   accepting tcp connections.  It doesn't take any parameters.  It returns a
   socket number for the newly created socket.  If any system call fails, it
   returns esmFAILURE.  This function is used by the supervisor process.
   The parameter supervisor port is used to determine in which port the
   supervisor will listen to.
 */


int create_passive_tcp_socket(int supervisor_port)
{
	
        int sock;       /* sock descriptor returned by the socket sys. call */
	struct sockaddr_in supervisor; /* Address in Internet domain */
        char *hostname = NULL;    /* pointer returned by gethostname */



	/* Create a socket */

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
		perror("socket");
		return(esmFAILURE);
	}
	
	/* Initialize the address structure for binding to a port */

        bzero((char *) &supervisor, sizeof supervisor);
	supervisor.sin_family = AF_INET;          /* Internet family type */
	supervisor.sin_port = htons( (short) supervisor_port ); /* assigning the port */
	supervisor.sin_addr.s_addr = INADDR_ANY;  /* Using zero host address so 
						 that local host address will 
						 be automatically supplied */

	/* Bind the socket to above port */

	if (bind(sock,(struct sockaddr *)&supervisor, sizeof supervisor) < 0) {
		perror("bind");
		return(esmFAILURE);
	}

	/* Get the name of the host on which the server is running */

	hostname = (char *)malloc(MAXHOSTNAMELEN);
	if (gethostname((char *)hostname, MAXHOSTNAMELEN) < 0) {
		perror("gethostname");
		return(esmFAILURE);
	}

	/* Print the name of the host on which the server is running */

	fprintf(stdout,"Server started on %s, ",hostname);
	fprintf(stdout,"port number %d\n\n", ntohs(supervisor.sin_port));

	/* Free the space allocated for holding the host name */

	free(hostname);

	/* Do a listen as the first step to become a passive socket */

	if (listen(sock, SOMAXCONN) == -1) {
		perror("listen");
		return(esmFAILURE);
	}

	/* Return the passive TCP socket */

	return(sock);
}

/* ========================================================================= */

/*  This function is used to accept a new connection on the given passive 
    socket.  It returns the socket number of the new connection, which is the
    return value of the accept system call.  This function is used by the 
    supervisor process.  If any system call fails, it returns esmFAILURE.*/


int accept_new_connection(
	int sock 	/* Socket on which to accept a new connection */
)
{
	int new_socket;  /* The socket number assigned to the new connection */
        struct sockaddr_in from; /* Address in Internet domain. Struct used to
				    hold the address of the host from which a
				    connection request came in */
	int fromlen = sizeof (struct sockaddr);
	struct hostent *hostname; /* structure returned by gethostbyaddr */


	/* Receive a connection request from the port */

	if ((new_socket = accept(sock, (struct sockaddr *) &from, &fromlen))
								< 0)  {
		perror("accept");
		return(esmFAILURE);
	}

	/* Get name and port no. of worker from which request was received */

	hostname = gethostbyaddr((char *) &from.sin_addr.s_addr,
				sizeof (from .sin_addr.s_addr), AF_INET);
	if (hostname == NULL) {
		perror("gethostbyaddr");
		return(esmFAILURE);
	}

#ifdef DEBUG
	fprintf(stdout,"SERVER :");
	fprintf(stdout,"Received connect request from MR \"%s",hostname->h_name);
	fprintf(stdout," : port no. %hd\"\n", ntohs(from.sin_port));
#endif

	/* Return the socket number for the new connection */

	return(new_socket);
}

/* ========================================================================= */

/*  This function is used to create an active TCP socket, which can be used
    either to connect to a passive socket or to another active socket.  This 
    function takes no input parameters, but takes an output parameter which is
    a pointer to an int, in which the port number of the newly created socket
    is returned.  The return value of the funtion is the socket number of the
    newly created socket.  If any system call fails, it returns esmFAILURE.*/

int create_active_tcp_socket(
	int *port_number	
		/* output parameter used to pass the port number to
			   which the newly created socket is bound to */
)
{
	int sock;	/* The socket number of the newly created socket */
	struct sockaddr_in worker; 	/* Addr. struct. in Internet */
	struct sockaddr_in name;/* Addr. struct. in Internet used in getting
				   the port number of the worker */
	int namelen = sizeof (struct sockaddr); 
        char *hostname = NULL;  /* char ptr. to the name of the host */


	/* Create a socket */

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		return(esmFAILURE);
	}

	/* Initialize the address struct. to bind to a port */

	bzero((char *) &worker, sizeof worker);
	worker.sin_family = AF_INET;
	worker.sin_port = 0;
	worker.sin_addr.s_addr = INADDR_ANY; 

	/* Bind the socket to a free port */

	if (bind(sock, (struct sockaddr *) &worker, sizeof worker) < 0) {
		perror("bind");
		return(esmFAILURE);
	}

	/* Get the name of the host on which the worker is running */

	hostname = (char *)malloc(MAXHOSTNAMELEN);
	if (gethostname(hostname, MAXHOSTNAMELEN) < 0) {
		perror("gethostname");
		return(esmFAILURE);
	}

	/* Get the port number that the worker is bound to */

	if (getsockname(sock, (struct sockaddr *) &name, &namelen) < 0) {
		perror("getsockname");
		return(esmFAILURE);
	}

	*port_number =  (int)((unsigned int) ntohs(name.sin_port));

	/* Print the info. obtained above */

#ifdef DEBUG
	fprintf(stdout,"MR : ");
	fprintf(stdout,"New port created on %s ",hostname);
	fprintf(stdout,"; port number is %d\n\n", *port_number);
#endif
	
	/* Return the socket number of the newly created socket */

	return(sock);
}

/* ========================================================================= */

/* This function is used to connect to the given port number on the given 
   machine using the given socket in the local process.  It takes the following
   input parameters: Socket number in the local process to be used to set up
   the connection; the 32-bit internet address of the machine to which to 
   connect to; the port number on the above machine to which to connect to.
   It returns esmNOERROR on success and esmFAILURE on error */


int connect_to_given_port(
	int sock,		/* Socket in the local process to be used to set up conn. */
	u_long host_addr,	/* 32-bit internet address of machine to connect to */
	int port_number		/* port_number to connect to on the above machine */
)
{
        struct sockaddr_in dest;	/* Addr. struct. in Internet */

#ifdef DEBUG
fprintf(stdout,"Connecting to host addr %x and port %d\n", host_addr, port_number);
#endif

	/* Initialize destination address struct. to prepare for a connect */
	
	bzero((char *) &dest, sizeof dest);
	dest.sin_family = AF_INET;
	dest.sin_port = htons((short)port_number);

	BCOPY((char *)&host_addr, (char *) &dest.sin_addr, sizeof host_addr);
	
	/* Connect to the given machine */

	if (connect(sock, (struct sockaddr *)&dest, sizeof(struct sockaddr))
									< 0) {
		perror("connect");
		fprintf(stdout,"errno = %d\n", errno);
		return(esmFAILURE);
	}

		return(esmNOERROR);
}

/* ========================================================================= */
#endif MEDIA_RECOVERY
