#ifdef MEDIA_RECOVERY
/*
 *   $RCSfile: mr_CopyDataPage.c,v $  
 *   $Revision: 1.7 $  
 *   $Date: 1993/02/04 20:30:06 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "pool.h"
#include "io_extfuncs.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "log_globals.h"
#include "bf_globals.h"
#include "mr_extfuncs.h"


 int
mr_CopyDataPage (

	PID     	pid,
	int      	*length,
 	GROUPLINK 	**returnLink 	
)
{

	VOLREC		*volRec;
	GROUPLINK	*groupLink;
	PAGEHASH	*pageHash;
	
	TRACE(TR_LOG, TR_LEVEL_1);

	volRec = io_FindVolRec(pid.volid);
	if (volRec == NULL) {
		SM_ERROR(TYPE_USER, esmBADVOLID);
		return(esmFAILURE);
	}

    /*
     *  check to see if the bit address is valid
     */
    if ((pid.page < 0) || (pid.page > volRec->header->numPages))  {

        SM_ERROR(TYPE_USER, esmBADPID);
        return(esmFAILURE);
    }

	/*
	 *	Read the page *from disk*.  We don't want a copy from the
	 *	buffer that is being dirtied
	 */

	/*
	 *	manage the buffer group space
	 */
	if (BF_Swap(UserBufGroup, MIN_PAGE2SIZE))   {
		SM_ERROR(TYPE_CRASH, esmINTERNAL);
	}

	/*
	 *	Get a buffer
	 */
	if ((groupLink = bf_GetPageBuf(UserBufGroup, MIN_PAGE2SIZE)) == NULL) {
		SM_ERROR(TYPE_CRASH, esmINTERNAL);
	}

	/*
	 *  read the page
	 */
	if (io_ReadPage(&pid, groupLink->pageHash->bufFrame, MIN_PAGE2SIZE)) {
		SM_ASSERT(LEVEL_1, LIST_EMPTY(&(pageHash->semaphore.waitList)));

	}

/* 
 *	Do this after the page is sent
 *
	signalSemaphore( &(groupLink->pageHash->semaphore) );
	bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
 */

	/*
	 *	Return the length of the page read
	 *	and the page with the data
	 */
	*length = MIN_PAGESIZE;
	*returnLink = groupLink;
	return(esmNOERROR);
}
#endif MEDIA_RECOVERY
