/*
 *   $RCSfile: redoConvertToLargeRoot.c,v $  
 *   $Revision: 1.10 $  
 *   $Date: 1993/03/08 13:42:44 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "pool.h"
#include "volume.h"
#include "logrecs.h"
#include "trans.h"
#include "openlog.h"
#include "bitmap.h"
#include "bf_extfuncs.h"
#include "sm_extfuncs.h"
#include "bm_extfuncs.h"
#include "trans_extfuncs.h"
#include "redo_intfuncs.h"
#include "redo_extfuncs.h"
#include "logaction.h"
#include "util_funcs.h"
#include "thread_globals.h"
#include "bf_globals.h"
#include "log_globals.h"
#include "log_extfuncs.h"
#include "sm_macro.h"
#include "bf_macro.h"


 void
redoConvertToLargeRoot (

	LOGRECORDHDR		*record  
)
{ 
	PID					*pid; 
	GROUPLINK			*groupLink;
	SLOTTEDPAGE			*slottedPage;
	SMALLOBJ			*logObj;
	SMALLOBJ			*origObj;
	OID					*oid;
	PAGESLOT			*slotPtr;

	TRPRINT(TR_IO|TR_LOG, TR_LEVEL_1, ("lsn:%d", record->recordLSN.offset));

	/*
	 *	get a pointer to the page for the record
	 */
	pid = &(record->actionPid);
	TRPRINT(TR_IO|TR_LOG, TR_LEVEL_2, ("pid:%d", pid->page));

	/*
	 *	get pointer to oid,  and the new object data
	 */
	oid = (OID*) GET_LOG_IMAGE(record, 0);
	logObj = (SMALLOBJ*) GET_LOG_IMAGE(record, 1);
	SM_ASSERT(LEVEL_3, GET_LOG_IMAGE_SIZE(record, 1) == 
						sizeof(LARGEOBJROOT));
					   

    /*
     *  See if the operation needs to be redone
     */
    if (!redoCheckPage(record, &groupLink, PAGE_SLOTTED)) {

        /*
         *  don't need to redo
         */
        TRPRINT(TR_RECOVER, TR_LEVEL_2, ("dirty page not present"));
        return;
    }
    CHECK_GROUPLINK_MAGIC(groupLink);

	/*
	 *	get a pointer to the slotted page
	 */
	slottedPage = (SLOTTEDPAGE *) groupLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(slottedPage)) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Get a pointer to the slot and then the object
	 */
	SM_ASSERT(LEVEL_3, oid->diskAddr.slot < slottedPage->header.slotCount);
	slotPtr = GETSLOTPTR(slottedPage, oid->diskAddr.slot);
	SM_ASSERT(LEVEL_3, oid->diskAddr.unique == slotPtr->unique);
	origObj  = GETOBJECTPTR(slottedPage, slotPtr);

	/*
	 *	The object should have no data in it.  
	 */
	SM_ASSERT(LEVEL_3, (origObj->objectHdr.size == 0) || (origObj->objectHdr.properties & P_MOVED));

	/* 
	 *	Redo the conversion by inserting the slotted root and
	 *	updating the object header.  This will properly handle the
	 *	case were the object is moved.
	 */
	if (sm_AppendToSmallObj(slottedPage, logObj->data, oid, sizeof(LGSLOTROOT)))	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
#ifdef JUNK
	/*
	 *  If the original object was forwarded space is freed up on the page.
	 */
	if (origObj->objectHdr.properties & P_MOVED) {

		/* now the forward oid is no longer stored */
		slottedPage->header.freeBytes += sizeof(OID);
	}
#endif JUNK
	origObj  = GETOBJECTPTR(slottedPage, slotPtr);
	origObj->objectHdr = logObj->objectHdr;
	slottedPage->header.largeCount++;

	/*
	 *	mark the lrc on the page
	 */
	slottedPage->header.lrc = record->actionLRC;
	TRPRINT(TR_RECOVER, TR_LEVEL_2, ("marking new page lrc:%d, ", slottedPage->header.lrc.count));

	/*
	 *  validate the page
	 */
	sm_AuditSlotPage(slottedPage);

	/*
	 *	signal the semaphore
	 */
	signalSemaphore( &(groupLink->pageHash->semaphore) );

	/*
	 *	unfix the page
	 */
	bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
}
