/*
 *   $RCSfile: redoInitLargeNode.c,v $  
 *   $Revision: 1.9 $  
 *   $Date: 1992/11/19 06:09:21 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "pool.h"
#include "volume.h"
#include "logrecs.h"
#include "trans.h"
#include "openlog.h"
#include "bitmap.h"
#include "bf_extfuncs.h"
#include "redo_extfuncs.h"
#include "redo_intfuncs.h"
#include "logaction.h"
#include "util_funcs.h"
#include "thread_globals.h"
#include "bf_globals.h"
#include "log_globals.h"
#include "sm_macro.h"
#include "bf_macro.h"


/*
 *	initialize a large object index node
 */

 void
redoInitLargeNode (

	LOGRECORDHDR		*record 
)
{

	PID					*pid;
	GROUPLINK			*groupLink;
	LGNODE				*lgNode;
	LGNODEHDR			*lgNodeHdr;
	SLOTINDEX			rootSlot;

	TRPRINT(TR_IO|TR_LOG, TR_LEVEL_1, ("lsn:%d", record->recordLSN.offset));

	/*
	 *	get a pointer to the page for the record
	 */
	pid = &(record->actionPid);
	TRPRINT(TR_IO|TR_LOG, TR_LEVEL_2, ("pid:%d", pid->page));

	/*
	 *	get pointer to oid, header info
	 */
	rootSlot = *(SLOTINDEX*) GET_LOG_IMAGE(record, 0);

	/*
	 *	get a pointer to the new data and its length
	 */
	lgNodeHdr = (LGNODEHDR*) GET_LOG_IMAGE(record, 1);
	CHECK_LARGENODE_MAGIC((LGNODE*)lgNodeHdr);

	/*
	 *	See if the operation needs to be redone
	 *	The assert guarantees the operation is never called on
	 *	a LARGESLOTROOT.
	 */
	SM_ASSERT(LEVEL_3, rootSlot == NIL); 
	if (!redoCheckPage(record, &groupLink, PAGE_LARGENODE)) {

		/*
		 *	don't need to redo
		 */
		TRPRINT(TR_RECOVER, TR_LEVEL_2, ("dirty page not present"));
		return;
	}
	CHECK_GROUPLINK_MAGIC(groupLink);

	/*
	 *	get a pointer to the large data page
	 */
	lgNode = (LGNODE *) groupLink->bufFrame;
	CHECK_LARGENODE_MAGIC(lgNode);

#ifdef DEBUG
	PRINT_PROGRESS(("redo write to page:%d\n", pid->page));
#endif	 

	/*
	 *	Perform the redo 
	 */
	lgNode->header = *lgNodeHdr;

	/*
	 *	mark the lrc on the page
	 */
	lgNode->header.lrc = record->actionLRC;
	TRPRINT(TR_RECOVER, TR_LEVEL_2, ("marking new page lrc:%d, ", lgNode->header.lrc.count));

	/*
	 *	signal the semaphore
	 */
	signalSemaphore( &(groupLink->pageHash->semaphore) );

	/*
	 *	unfix the page
	 */
	bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
}
