/*
 *   $RCSfile: redoInsertLargeData.c,v $  
 *   $Revision: 1.13 $  
 *   $Date: 1993/06/21 12:51:01 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "pool.h"
#include "volume.h"
#include "logrecs.h"
#include "trans.h"
#include "openlog.h"
#include "bitmap.h"
#include "bf_extfuncs.h"
#include "redo_extfuncs.h"
#include "redo_intfuncs.h"
#include "logaction.h"
#include "util_funcs.h"
#include "thread_globals.h"
#include "bf_globals.h"
#include "log_globals.h"
#include "sm_macro.h"
#include "bf_macro.h"


/*
 *	Reinsert data into a page 
 */

 void
redoInsertLargeData (

	LOGRECORDHDR		*record 
)
{

	PID					*pid;
	GROUPLINK			*groupLink;
	LGDATAPAGE			*lgDataPage;
	char				*data;
	int					dataLength;
	LGWRITEINFO			*lgWriteInfo;
	int					i;

	TRPRINT(TR_IO|TR_LOG, TR_LEVEL_1, ("lsn:%d", record->recordLSN.offset));

	/* get a pointer to the page for the record */
	pid = &(record->actionPid);
	TRPRINT(TR_IO|TR_LOG, TR_LEVEL_2, ("pid:%d", pid->page));

	/* get pointer to oid, header info */
	lgWriteInfo = (LGWRITEINFO*) GET_LOG_IMAGE(record, 0);

	/*
	 *	get a pointer to the new data and its length
	 */
	dataLength = lgWriteInfo->size;
	switch (record->action) {
	case LOG_ACTION_INSERT_LARGE_DATA:
		data = GET_LOG_IMAGE(record, 1);
		break;
	case LOG_ACTION_DELETE_LARGE_DATA:
	case LOG_ACTION_INSERT_LARGE_DATA_ND:
		data = NULL;
		break;
	default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		break;
	}

	/*
	 *	See if the operation needs to be redone
	 */
	if (!redoCheckPage(record, &groupLink, PAGE_LARGEDATA)) {
		/* don't need to redo */
		TRPRINT(TR_RECOVER, TR_LEVEL_2, ("dirty page not present"));
		return;
	}
	CHECK_GROUPLINK_MAGIC(groupLink);

	/* get a pointer to the large data page */
	lgDataPage = (LGDATAPAGE *) groupLink->bufFrame;
	CHECK_LGDATAHDR_MAGIC(lgDataPage);

#ifdef DEBUG
	PRINT_PROGRESS(("redo insert to large page:%d\n", pid->page));
#endif	 

	if (record->action == LOG_ACTION_INSERT_LARGE_DATA ||
		record->action == LOG_ACTION_INSERT_LARGE_DATA_ND) {

		/*
		 *	Make a hole for the new data and copy it in
		 *	The for loop i > lg... comparison cannot be >=
		 *	since i is cast to unsigned for the comparison.
		 */
		for (i = lgWriteInfo->prevSize; i > lgWriteInfo->start; i--) {
			lgDataPage->data[(i-1)+dataLength] = lgDataPage->data[i-1];
		}
		if (data != NULL) {
			bcopy(data, lgDataPage->data + lgWriteInfo->start, dataLength);
		} else {
			bzero(lgDataPage->data + lgWriteInfo->start, dataLength);
		}

		/* make sure magic did not get messed up */
		CHECK_LGDATAHDR_MAGIC(lgDataPage);

	} else {

		SM_ASSERT(LEVEL_3, record->action == LOG_ACTION_DELETE_LARGE_DATA);

		/*
		 *	Shift data down to delete 
		 */
		for (i = lgWriteInfo->start+dataLength; i < lgWriteInfo->prevSize+dataLength; i++) {
			lgDataPage->data[i-dataLength] = lgDataPage->data[i];
		}

		/* make sure magic did not get messed up */
		CHECK_LGDATAHDR_MAGIC(lgDataPage);
	}

	/* mark the lrc on the page */
	lgDataPage->header.lrc = record->actionLRC;
	TRPRINT(TR_RECOVER, TR_LEVEL_2, ("marking new page lrc:%d, ", lgDataPage->header.lrc.count));

	/* release the semaphore */
	signalSemaphore( &(groupLink->pageHash->semaphore) );

	bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
}
