/*
 *   $RCSfile: sm_AuditSlotPage.c,v $  
 *   $Revision: 2.22 $  
 *   $Date: 1993/04/14 21:37:20 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#ifdef SERVER_MAKE
#   include "msgdefs.h"
#   include "lsn.h"
#   include "link.h"
#   include "thread.h"
#   include "semaphore.h"
#   include "volume.h"
#   include "trans.h"
#   include "latch.h"
#endif
#include "bf.h"
#ifndef SERVER_MAKE
#   include "chunk.h"
#endif
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#ifndef SERVER_MAKE
#   include "scan.h"
#   include "load.h"
#include "sm_params.h"
#include "btree.h"
#endif
#include "trace.h"
#include "sm_extfuncs.h"

#ifdef SERVER_MAKE
#   include "thread_globals.h"
#endif SERVER_MAKE


/*
 *	Here we check the sizes of the basic datatypes used in
 *	slotted pages to make sure they are all multiple of the
 *	object alignment macro ALIGNON (usually 8).  Hopefully,
 *	the compiler's constant folding will report problems
 *	immedidately.
 */
#ifdef DEBUG
static int dummy1 = 1/ ( (sizeof(OBJHDR) % ALIGNON) == 0 ? 1: 0);
static int dummy2 = 1/ ( (sizeof(LARGEOBJROOT) % ALIGNON) == 0 ? 1: 0);
static int dummy3 = 1/ ( (sizeof(LARGEOBJHDR) % ALIGNON) == 0 ? 1: 0);

 void
SM_AuditSlotPage (

	register SLOTTEDPAGE	*page ,
	char					*file,
	int						line

)
{
	static 					int	 nth_call = 0;

    register OBJECT			*objectPtr;		/* ptr to generic object		*/
	PAGESLOT				*slotPtr;
	SLOTINDEX				i, curr;
	int						largeCount;
	int						usedSpace, slotUsedSpace;
	LARGEOBJROOT 			*lobjRoot = NULL;
	LARGEOBJHDR 			*lobjHdr = NULL;
	int						slotsUsed = 0;
	int						slotsFree = 0;

	TRPRINT(TR_SM|TR_EH, TR_LEVEL_1, 
		("auditing slot page:%d: freeStart %d, freeBytes %d",
		page->header.pid.page, page->header.freeStart,
		page->header.freeBytes));
	TRPRINT(TR_EH, TR_LEVEL_1, ("called from %s line %d", file, line));

	nth_call++;

	TRPRINT(TR_SM|TR_EH, TR_LEVEL_1, 
		("total slots %d, first free slot %d", 
		page->header.slotCount, page->header.freeSlot));

	/*
	 *	check to make sure the page magic is correct
	 */
	if (CHECK_SLOTTED_MAGIC(page))	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	check out all the slots and the object headers
	 *	calculate all the space used by objects
	 */
	largeCount = 0;
	usedSpace = 0;
	for (i = 0; i < page->header.slotCount; i++)	{

		TRPRINT(TR_EH, TR_LEVEL_1, ("slot:%d", i));

		/*
		 *	get a pointer to the slot
		 */
		slotPtr = GETSLOTPTR(page, i);

		/*
		 *	check to see if slot is valid
		 */
		if (slotPtr->offset != NIL)	{

			TRPRINT(TR_EH, TR_LEVEL_1, ("slot %d is valid", i));

			slotsUsed++;

			/* make sure obj begins on 8byte boundry */
			SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);

			/*
			 *	get a pointer to the object
			 */
			objectPtr = GETOBJECTPTR(page, slotPtr);


#if OBJECT_MAGIC IS_ENABLED

			/*
			 *	check out the magic number in the object header
			 */
			if (objectPtr->objectHdr.magic != OBJECT_HDR_MAGIC)	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}

#endif
			if (objectPtr->objectHdr.properties & P_LARGEOBJ) {
				TRPRINT(TR_EH, TR_LEVEL_1, ("slot %d is large object", i));
				largeCount++;
			}

			/*
			 *	Calculate the space used by the obj
			 */
			if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

				if (objectPtr->objectHdr.properties & P_HEADERONPAGE)   {
					slotUsedSpace = ALIGNSIZE(sizeof(LARGEOBJROOT));
					{
						register int j, sum;

						lobjRoot = (LARGEOBJROOT *) objectPtr;

						TRPRINT(TR_EH, TR_LEVEL_1, 
						("large obj, hdr on page, size %d, props %x %d slots", 
							objectPtr->objectHdr.size, 
							objectPtr->objectHdr.properties,
							lobjRoot->rootPage.header.numSlots));

						if( (lobjRoot->rootPage.header.numSlots > 0) &&
							(objectPtr->objectHdr.size != 
							lobjRoot->rootPage.slot[
								lobjRoot->rootPage.header.numSlots-1 ].count) ) {
							TRPRINT(TR_EH, TR_LEVEL_1, 
								("objectHdr size %d, rootPage size %d",
								objectPtr->objectHdr.size,
								lobjRoot->rootPage.slot[
								lobjRoot->rootPage.header.numSlots-1 ].count));
#ifdef SERVER_MAKE
							if( (Active->transRec != NULL) &&
								(((TRANSREC *)Active->transRec)->transState ==
									T_ACTIVE) ) {
									/* not in recovery or abort - all
									 * pages should be consistent 
									 */
#endif SERVER_MAKE
							SM_ERROR(TYPE_FATAL, esmINTERNAL);
#ifdef SERVER_MAKE
							} else {
								if (Active->transRec != NULL) {
								TRPRINT(TR_EH, TR_LEVEL_1, 
						("TODO: IS THIS AN ERROR???????? (trans state %d)",
								((TRANSREC *)Active->transRec)->transState ));
								}
								else {
								TRPRINT(TR_EH, TR_LEVEL_1, 
						("TODO: IS THIS AN ERROR???????? (no transRec)"));
								}

							}
#endif SERVER_MAKE
						}
						for( (sum = lobjRoot->rootPage.slot[0].count),
							j = (int) lobjRoot->rootPage.header.numSlots -1; 
							j > 0;
							j--) {

							TRPRINT(TR_EH, TR_LEVEL_1, 
							("large obj hdr slot %d, page %d, count %d", 
								j,
								lobjRoot->rootPage.slot[j].pid.page,
								lobjRoot->rootPage.slot[j].count));
							sum += (int) (lobjRoot->rootPage.slot[j].count -
								lobjRoot->rootPage.slot[j-1].count);
						}

						if(objectPtr->objectHdr.size != sum) {
#ifdef SERVER_MAKE
							if( (Active->transRec != NULL) &&
								(((TRANSREC *)Active->transRec)->transState ==
									T_ACTIVE) ) {
									/* not in recovery or abort - all
									 * pages should be consistent 
									 */
#endif SERVER_MAKE
								SM_ERROR(TYPE_FATAL, esmINTERNAL);
#ifdef SERVER_MAKE
							} else {
								if (Active->transRec != NULL) {
								TRPRINT(TR_EH, TR_LEVEL_1, 
						("TODO: IS THIS AN ERROR???????? (trans state %d)",
								((TRANSREC *)Active->transRec)->transState ));
								}
								else {
								TRPRINT(TR_EH, TR_LEVEL_1, 
						("TODO: IS THIS AN ERROR???????? (no transRec)"));
								}

							}
#endif SERVER_MAKE
						}
					}
				} else {
					slotUsedSpace = ALIGNSIZE(sizeof(LARGEOBJHDR));
					lobjHdr = (LARGEOBJHDR *) objectPtr;
					if (lobjHdr->rootPid <= BITMAPADDR) {
						/* illegal page id */
						SM_ERROR(TYPE_FATAL, esmINTERNAL);
					}
				}
			} else {
				if (objectPtr->objectHdr.properties & P_MOVED)  {
					slotUsedSpace = sizeof(OBJHDR) + ALIGNSIZE(sizeof(OID));
				} else {
					slotUsedSpace = (int)(sizeof(OBJHDR) + ALIGNSIZE(objectPtr->objectHdr.size));
				}
			}
			usedSpace += slotUsedSpace;
			TRPRINT(TR_EH, TR_LEVEL_1, 
				("slot %d uses %d bytes", i, slotUsedSpace));

		} else {

			TRPRINT(TR_EH, TR_LEVEL_1, ("slot %d is not used", i));
			slotsFree++;
		}
	}

	if ( (slotsUsed+slotsFree) != page->header.slotCount) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Check the free slot list, by traversing an subtracting 1
	 *	from slotsFree for each free slot.  Result should be zero.
	 */
	curr = page->header.freeSlot;
	while (curr != NIL) {
		SM_ASSERT(LEVEL_1, curr < page->header.slotCount);
		slotPtr = GETSLOTPTR(page, curr);
		SM_ASSERT(LEVEL_1, slotPtr->offset == NIL);
		slotsFree--;

		/* go to the next slot */
		curr = (SLOTINDEX) slotPtr->unique;
	}
	if (slotsFree != 0) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	if (largeCount != page->header.largeCount) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	add the space used by the slots and the page header
	 */
	usedSpace += page->header.slotCount * sizeof(PAGESLOT);
	usedSpace += sizeof(page->header);


	SM_ASSERT(LEVEL_3, page->header.freeBytes >= 0);

	if ( page->header.freeBytes != (sizeof(SLOTTEDPAGE) - usedSpace) ) {
		TRPRINT(TR_SM|TR_EH, TR_LEVEL_1, ("freeBytes is %d, should be %d ", 
			page->header.freeBytes, (sizeof(SLOTTEDPAGE) - usedSpace)));
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
}

#endif
