/*
 *   $RCSfile: asyncAbort.c,v $  
 *   $Revision: 2.17 $  
 *   $Date: 1993/02/11 19:49:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "latch.h"
#include "bf.h"
#include "volume.h"
#include "bitvec.h"
#include "trans.h"
#include "openlog.h"
#include "logrecs.h"
#include "log_globals.h"
#include "trans_intfuncs.h"
#include "trans_extfuncs.h"
#include "thread_funcs.h"
#include "lm_extfuncs.h"
#include "msg_funcs.h"
#include "log_extfuncs.h"
#include "thread_globals.h"
#include "trans_globals.h"

#ifdef vax
 void
asyncAbort (
    register TRANSREC   *transRec,
    register int	   	reason
)
#elif defined(mips) || defined(sparc) || defined(_AIX) || defined(hpux)
 void
asyncAbort ()
#elif !(defined(mips) || defined(sparc) || defined(vax) || defined(_AIX) || defined(hpux))
    not supported
#else
    /* this else should have the not supported code, but the
       ultrix cpp is broken */
#endif

{
	/*
	 *	This variable holds an argument passed the this function
	 *	when the thread was forked.	 This should be portable across
	 *	all architectures but vax's.
	 */
#if defined(mips) || defined(sparc) || defined(_AIX) || defined(hpux)
    TRANSREC	*transRec;
    int			reason;
#elif defined(vax)
	/* not needed */
#elif !(defined(mips) || defined(sparc) || defined(vax) || defined(_AIX) || defined(hpux))
    not supported
#else
    /* this else should have the not supported code, but the
       ultrix cpp is broken */
#endif 

	TRACE(TR_TRANS, TR_LEVEL_1);

	/*
	 *	suck the transrec out of the tcb buffer
	 */
#ifdef mips
/*
    transRec = (TRANSREC *) Active->currentBuf[JB_A0];
    reason = Active->currentBuf[JB_A0+1];
*/
    transRec = (TRANSREC *) Active->args[0];
    reason = Active->args[1];

#elif defined(sparc) || defined(_AIX) || defined(hpux)
	/*
	 *	Sparc and other architectures
	 */
    transRec = (TRANSREC *) Active->args[0];
    reason = Active->args[1];
#elif defined(vax)
	/* not needed */
#elif !(defined(mips) || defined(sparc) || defined(vax) || defined(_AIX) || defined(hpux))
    not supported
#else
    /* this else should have the not supported code, but the
       ultrix cpp is broken */
#endif 

	TRPRINT(TR_TRANS, TR_LEVEL_2, ("tid:%d", GETTID(transRec)));

	/*
	 *	mark the active transaction record
	 */
	Active->transRec = (char *) transRec;

	/*
	 *	check the transaction magic
	 */
	CHECK_TRANSREC_MAGIC(transRec);

	/*
	 *	check to see if there is a transaction master
	 */
	if (transRec->masterThread != NULL)	{

		TRPRINT(TR_TRANS, TR_LEVEL_1, ("tid:%x has master", GETTID(transRec)));
		threadRestart();
	}

	/*
	 *	check to see if the transaction is active
	 */
	if ((transRec->transState != T_ACTIVE) &&
		(transRec->transState != T_LATENT))	{

		SM_ERROR(TYPE_LOG, esmTRANSABORTED);
		threadRestart();
	}

	/*
	 *	mark the intent to commit
	 */
	transRec->intent = T_ABORT;

	/*
	 *	quiesce all resources held by the transaction
	 */
	quiesceTrans(transRec, esmTRANSABORTED);

	/*
	 *	Finish aborting the transaction
	 */
	abortTrans(transRec, reason);

	/*
	 *	restart this task
	 */
	threadRestart();
}
