/*
 *   $RCSfile: LG_AllocNodes.c,v $  
 *   $Revision: 2.13 $  
 *   $Date: 1993/03/23 23:38:28 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"


/*
 * Routine to allocate a new internal node due to overflow, and to add the
 * pid of the newly allocated page to the new page list for subsequent use.
 */


 int 
LG_AllocNodes (
	BUFGROUP			*bufGroup,		/* buffer group for operation		*/
	CHUNKPAGE			*chunkPage,		/* list of pid/cnt pairs to insert	*/
	PID					*nearPid,		/* pid to allocate near				*/
	int					count 			/* number of nodes to allocate		*/
)
{
	register PID		*pidList;
	register GROUPLINK	*nodeLink = NULL;
	int		pagesFixedIndex = -1; /* don't use for anything else */


	TRPRINT(TR_LG, TR_LEVEL_1, ("count:%d", count));

	if ((pidList = (PID *) malloc(sizeof(PID) * count)) == NULL)	{
		SM_ERROR(TYPE_USER, esmMALLOCFAILED);
		FAIL
	}


    if (io_AllocPages(nearPid->volid, LG_PAGE2SIZE, count,
					  nearPid, pidList, PAGE_LARGENODE))	{
		FAIL
	}
	TRPRINT(TR_LG, TR_LEVEL_2, ("allocated %d starting at page:%d", count, pidList->page));

	/*
	 *	set up buffers to hold the new pages
	 */
	for (pagesFixedIndex = 0; pagesFixedIndex < count; 
		pagesFixedIndex++, chunkPage++)	{

#ifdef DEBUG
		{
			PAGEHASH *ph;
			ph  = bf_FindPage( &(pidList[pagesFixedIndex]) );
			SM_ASSERT(LEVEL_1, (ph == NULL) || (ph->flags & H_INVALID));
		}
#endif DEBUG

		/*
		 *	get a buffer to hold the new node
		 */
		nodeLink = bf_GetPage(bufGroup, &(pidList[pagesFixedIndex]), H_LARGE, NULL, LG_PAGE2SIZE);

		if (nodeLink == NULL)	{
			FAIL
		}

		/*
		 *	record the page information in the chunk entry
		 */
		chunkPage->pid = pidList[pagesFixedIndex];
		chunkPage->nodeLink = nodeLink;

		nodeLink->pageHash->flags |= H_NEW;
		bf_DirtyPage(nodeLink->pageHash);
	}
Success:

	SM_FREE_PTR(pidList);
    return esmNOERROR;

FailureReturn:
	for (; pagesFixedIndex >=0 ; pagesFixedIndex--, chunkPage--) {
		nodeLink = chunkPage->nodeLink; 
		if(nodeLink) {
			bf_UnfixPage(nodeLink,BF_DEFAULT, FALSE);
			chunkPage->nodeLink = NULL;
		}
		
	}
	SM_FREE_PTR(pidList);
    return esmFAILURE;

}
