/*
 *   $RCSfile: SM_GetNextPrevOid.c,v $
 *   $Revision: 1.9 $  
 *   $Date: 1993/04/09 19:21:31 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "file.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "bf_extfuncs.h"
#include "fi_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#include "trans_funcs.h"
#include "bf_globals.h"


extern int			sm_errno;


 int
SM_GetNextPrevOid (
	int				groupIndex,
	OID				*baseOid,
	OID				*nextOid,
	OBJHDR			*objectHdr,
	BOOL			*eof,
	NEIGHBORFLAGS	direction
)
{
	GROUPLINK		*currentLink;
	BUFGROUP		*bufGroup;
	SLOTTEDPAGE		*currentPage;
	PAGESLOT		*slotPtr;
	PID				currentPid, nextPid;
	SLOTINDEX		currentSlot;
	int				e;
	FID				fid;
	BOOL			tryAgain = TRUE;
#ifdef DEBUG
	int				fixedCharges;
#endif DEBUG


	TRPRINT(TR_SM, TR_LEVEL_1, ("volid:%d page:%d slot:%d unique:%d",
			baseOid->diskAddr.volid, baseOid->diskAddr.page,
			baseOid->diskAddr.slot, baseOid->diskAddr.unique));

	if (checkTransState(baseOid->diskAddr.volid))	{
		return(esmFAILURE);
	}

    /*
     * extract the pid from the oid
     */
    currentPid.volid = baseOid->diskAddr.volid;
    currentPid.page = baseOid->diskAddr.page;

	bufGroup = &(BufGroupPtr[groupIndex]);	
#ifdef DEBUG
	fixedCharges = (int) bufGroup->fixedCharges;
	TRPRINT(TR_SM, TR_LEVEL_1, ("fixedCharges %d", bufGroup->fixedCharges));
#endif DEBUG

	if (CHECK_BUFGROUP_VALID(bufGroup))	{

		SM_ERROR(TYPE_USER, esmBADGROUP);
		return(esmFAILURE);
	}

	/* 
	 * We have to get the page and the neighbor page in 2 steps-
	 * grot- we have to read the page to get the fid, then
	 * use the fid to get the neighbor page.
	 */

again:

	if(  (currentLink = bf_ReadPage(bufGroup, &currentPid,  
		PAGE_SLOTTED, SLOTTED_PAGE2SIZE, NL)) == NULL) {
			return(esmFAILURE);
	}

	currentPage = (SLOTTEDPAGE *) currentLink->bufFrame;
	fid = currentPage->header.fid;

	if (CHECK_SLOTTED_MAGIC(currentPage))	{
		bf_UnfixPage(currentLink, BF_DEFAULT, FALSE);
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
		return(esmFAILURE);
	}
    /*
     * Make sure the slot given by the oid is valid
     */
    currentSlot = baseOid->diskAddr.slot;
	slotPtr = GETSLOTPTR(currentPage, currentSlot);

    if (CHECKOID(currentSlot, currentPage, slotPtr, baseOid->diskAddr.unique)){

		bf_UnfixPage(currentLink, BF_DEFAULT, FALSE);
		SM_ERROR(TYPE_WARNING, esmBADOID);
		return(esmFAILURE);
	}
	TRPRINT(TR_SM, TR_LEVEL_2, ("oid is valid offset is:%d", slotPtr->offset));


#if OBJECT_MAGIC IS_ENABLED
	{
		OBJECT			*objectPtr;
		objectPtr = GETOBJECTPTR(currentPage, slotPtr);

		/*
		 *	check the object header magic number
		 */
		if (objectPtr->objectHdr.magic != OBJECT_HDR_MAGIC)	{
			bf_UnfixPage(currentLink, BF_DEFAULT, FALSE);
			SM_ERROR(TYPE_WARNING, esmBADOBJECTMAGIC);
			return(esmFAILURE);
		}
	}
#endif
	TRPRINT(TR_SM, TR_LEVEL_1, ("fixedCharges %d", bufGroup->fixedCharges));

	if(tryAgain)
		nextPid.page = NULLPID;

	/* First, try without getting the next page. If it returns eof==true
	 * we'll get the next page and try again.
	 * This is a mighty kludge to avoid going to the server for
	 * the next page each time, and also to keep the code for snooping
	 * around the slots in one place, namely SM_GetOid().
	 */
	e  = SM_GetOid (bufGroup, currentLink, 
		currentSlot, nextOid, &nextPid, objectHdr, eof, direction);

	TRPRINT(TR_SM, TR_LEVEL_1, ("fixedCharges %d", bufGroup->fixedCharges));
#ifdef DEBUG
	SM_ASSERT(LEVEL_1, fixedCharges == bufGroup->fixedCharges);
#endif DEBUG
	TRPRINT(TR_SM, TR_LEVEL_1, ("e %d, eof %d, sm_errno", e, eof, sm_errno));
	if(e != esmNOERROR) return e;
	if( *eof == FALSE ) return e;

	/* *eof is TRUE, and the page was unfixed by SM_GetOid() */

	if( fi_NeighborPid( &currentPid, &fid, NULL, NULL, direction, &nextPid) ) {
		if (sm_errno == esmEMPTYFILE) {
			return(esmFAILURE);
		}
		if ((sm_errno != esmAFTERLASTPID) &&
			(sm_errno != esmBEFOREFIRSTPID) ) {
			return(esmFAILURE);
		}
		/* fi_NeighborPid does not change the value of nextPid
		 * in the error case.
		 * Drop through to SM_GetOid, which will compute
		 * and set eof.
		 */
	}

	/*
	 * currentLink is no longer fixed, so go try again.
	 */
	if(tryAgain) {
		tryAgain = FALSE;
		goto again;
	} else 
		return esmNOERROR;
}
