/*
 *   $RCSfile: cl_WriteLogPage.c,v $  
 *   $Revision: 1.23 $  
 *   $Date: 1993/04/16 15:43:07 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "threadstate.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "bitvec.h"
#include "consist.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "msg_funcs.h"
#include "bf_extfuncs.h"
#include "thread_funcs.h"
#include "cl_funcs.h"
#include "log_extfuncs.h"
#include "trans_extfuncs.h"
#include "server_util_funcs.h"
#include "thread_globals.h"
#include "log_globals.h"
#include "io_globals.h"


extern char			*ShmAddress;

 void
cl_WriteLogPage (

	register MESSAGE	*message,
	register LINK		*link 
)
{

	GROUPLINK	*tailLink;
	GROUPLINK	*tempLink;
	char		*buffer;
	PID			nextPid;
	SHORTPID	newTailPage;
	int			logCount;	/* number of log rec's in page 	*/
	TRANSREC	*transRec;
	LSN			firstLSN;
	LSN			lastLSN;
	LSNOFFSET	tailLSN;
	FORCEMARK	forceMark;	/* force mark for the log page	*/
	FORCEMARK	asyncForceMark;	/* force to here in background */
	int 		i;

	TRPRINT(TR_CL, TR_LEVEL_1, ("writing log pid"));

	PRINT_PROGRESS( ("received log page\n") )

	/*
	 *	suspend the link so that it does not get selected on
	 *	the basis of the data that is sent
	 */
	suspendLink(link);

	/*
	 *	Get a temporary buffer page
	 */
	SM_ASSERT(LEVEL_3, message->body.page.page2size = OpenLog.page2size);
	i = 0;
	while ( (tempLink = bf_GetTempPage(message->body.page.page2size)) == NULL) {
		i++;
		if (i > 10000000) {
			/* we are in a loop, so crash */
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/* let another thread run which may free up space */
		Sleep(SLEEP_DONT_SLEEP);
	}
	buffer = tempLink->bufFrame;

	/* 	resume the link so that the data can be read */
	resumeLink(link);

	/*
	 *	read the data into shared memory
	 */
	if (receive(link, buffer, RAISETWO(message->body.page.page2size)) < 0)	{
		/*	blow the thread away, the link is gone discard the page */
		bf_FreeTempPage(tempLink);
		threadRestart();
	}

	/*
	 * Suspend the link so it does not get selected for input.
	 * No other client requests are handled until the page is
	 * placed in the log.
	 * This is too restrictive since lock and page requests could
	 * be handled, but it is the simplest way to keep out
	 * aborts and any other things that must wait until this page
	 * is in the log.
	 */
	suspendLink(link);

	/*
	 *	attempt to find the transaction record, TRUE => writing log
	 */
	if ((transRec = registerTrans(message->header.params.in.tid, TRUE)) == NULL) {
		
        /*
         *  make sure that if there is a tran under this id that it
         *  is being aborted
         */
        if ((transRec = findTransRec(message->header.params.in.tid)) != NULL) {
            if (transRec->transState == T_ACTIVE) {
                SM_ERROR(TYPE_FATAL, esmINTERNAL);
            }
        }

		bf_FreeTempPage(tempLink);

		if (message->header.replyRequested) {
			reply(esmFAILURE, Active->errno, 1, FALSE);
		} 
		resumeLink(link);
		threadRestart();
	}

	/*
	 *	If logging is turned off for this transaction, just return
	 *	the page to the buffer pool i.e. do not put it in the log.
	 */
	if (transRec->loggingOn == FALSE) {
		bf_FreeTempPage(tempLink);
		if (message->header.replyRequested) {
			reply(esmNOERROR, esmNOERROR, 1, FALSE);
		} else {
			transThreadDeath(Active);
		}
		resumeLink(link);
		threadRestart();
	}

	/*
	 *	Get a new page in the log, and obtain a latch on the log
	 */
	if (reserveClientLogPage(transRec, &newTailPage) != esmNOERROR) {

		/*
		 *	free temp page and reply with error
		 *	
		 *	NOTE: the error may indicate that there is no free
		 *  log space.  In that case
		 *	this thread should "wait" and try again.  But, the
		 *	thread code doen't really support this, so we leave it
		 *	up to the client to try again.  In fact, the trans will
		 *	likely be aborted, so this is no big deal.
		 *	
		 *	For now we just abort the transaction immediately.
		 *	It would be best to put this thread on a waiting list
		 *	for more log space.  Then when log space is available
		 *	it could be restarted.
		 */
		bf_FreeTempPage(tempLink);

		if (message->header.replyRequested) {
			reply(esmFAILURE, Active->errno, 1, FALSE);
		} else {
			/* Error, so abort transaction */
			internalAbort(transRec, Active->errno);
			transThreadDeath(Active);
		}
		resumeLink(link);
		threadRestart();
	}

	/*
	 *	Obtain buffer space for the new page
	 */
	nextPid.volid = OpenLog.volid;
	nextPid.page = LOG_PAGE_TO_BLOCK(newTailPage, &OpenLog);
	tailLink = tempLink;
	/* remove any old version of the page */
	bf_InvalidatePage(&nextPid);
	/* put the page into the buffer pool hash table */
	if ((bf_ConvertTempToRealPage(OpenLog.writeGroup, tailLink, 
									&nextPid, BF_SEM)) != esmNOERROR) {
		SM_ERROR(TYPE_FATAL, Active->errno);
	}

	/*
	 *	mark that the buffer is a log buffer
	 */
	 PSET_PAGE_TYPE(tailLink->pageHash, PAGE_LOG);

#ifdef JUNK
	/*
	 *	Copy the log page to the log buffer area, and free the
	 *	temp page
	 */
	bcopy(tempLink->bufFrame, tailLink->bufFrame, OpenLog.pageSize);
	bf_FreeTempPage(tempLink);
#endif JUNK

	/*
	 *	Convert all LSN's on the client log page to be actual
	 *	LSN's for the log.  
	 */
	if (convertClientLogPage(tailLink, newTailPage, transRec, &firstLSN, &lastLSN, &tailLSN, &logCount) != esmNOERROR) {

		/*
		 *	signal the semaphore and unfix the page
		 *	and then return the error code
		 */
		signalSemaphore( &(tailLink->pageHash->semaphore) );
		bf_UnfixPage(tailLink, BF_DEFAULT, FALSE);
		signalLatch( &(OpenLog.logLatch) );
		signalSemaphore( &(OpenLog.writeSemaphore) );

		if (message->header.replyRequested) {
			reply(esmFAILURE, Active->errno, 1, FALSE);
		} else {
			/* Error, so abort transaction */
			internalAbort(transRec, Active->errno);
			transThreadDeath(Active);
		}
		resumeLink(link);
		threadRestart();
	}

	/*
	 *	Atomically make the new client log page the end of the
	 *	log.  The end of the log points to empty space on the
	 *	end of the client log page.  The log latch and semaphore are
	 *	also released.
	 */
	forceMark = updateEndOfLog(transRec, tailLink, newTailPage, 
							   tailLSN, &firstLSN, &lastLSN, logCount);

	/*
	 *	Record the forcemark for the log buffer page 
	 */
	tailLink->pageHash->forceMark = forceMark;		
	tailLink->pageHash->firstMark = forceMark;

	/*
	 *	Record the force mark for all subsequent dirty data
	 *	pages sent on this link.  Also, record the last LSN
	 *	generated for this link.
	 */
	asyncForceMark = link->lastForceMark;
	link->lastForceMark = forceMark;

	/* signal the semaphore and unfix the page, mark the page dirty */
	signalSemaphore( &(tailLink->pageHash->semaphore) );
	bf_UnfixPage(tailLink, BF_DEFAULT, TRUE);

	if (message->header.replyRequested) {
		reply(esmNOERROR, esmNOERROR, 1, FALSE);
	} else {
		/* thread is no longer running under a transaction */
		transThreadDeath(Active);
	}
	resumeLink(link);
	threadRestart();

	/* cannot get here */
	SM_ASSERT(LEVEL_3, esmINTERNAL);

}

