/*
 *   $RCSfile: thread.h,v $  
 *   $Revision: 2.26 $  
 *   $Date: 1993/04/26 21:56:42 $      
 */ 
#ifndef __THREAD_H__
#define __THREAD_H__

/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "ess.h"
#include "io.h"
#include "checking.h"
#include "msgdefs.h"
#include <setjmp.h>
#include "disk.h"

/*
 *	CONST_THREADS are always waiting to run.
 *	Currently there are :
 *		one for taking checkpoints
 *	 	one for flushing dirty pages after a checkpoint
 *	    one for log preflush
 *		one for coordinating distributed transactions
 *		one for aging locks
 *		one for timing out clients
 *		one for 2PC coordination
 */
#define CONST_THREADS 7

/*
 *	define some machine specific stack things
 */
#define FRAMEMARK			0xFFFFFFFF
#define USEDSTACK			9
#define STACKPATTERN		0xABCDABCD
#define THREAD_STACKSIZE	(0x800)

/*
 * NB: we use _longjmp because (on some machines) it does not do 
 * a stack "sanity" check that blows us away.
 * On the MIPS, this is utterly undocumented.
 * On the IBM, both _longjmp and longjmp do the sanity check, so we
 * have our own implementation of longjmp for AIX.
 * (We don't just define _longjmp for AIX, because we'd get a linkage
 * conflict if we did: setjmp/longjmp are in one .o file; _setjmp/_longjmp in another,
 * in general.)
 * On HP/UX and sparcs, all seems to be well with _longjmp from C-lib,
 * but nowhere is there a prototype for it!
 */
#ifndef _AIX
#define longjmp _longjmp
#endif

BEGIN_EXTERNC
	extern void longjmp( jmp_buf, int );
END_EXTERNC

#ifdef sparc
	/*
	 *	Location of pc and sp in sparc jump buf
	 */
#	define JB_SP_SPARC	2
#	define JB_PC_SPARC	3
#	define JB_NPC_SPARC	4

#endif sparc

#ifdef _AIX

    /*
     *  Location of pc and sp in RISC System/6000 jump buf
     */
#   define JB_SP_AIX    3
#   define JB_PC_AIX    2

/*
 * Structure needed to take a function pointer apart.
 */
typedef struct {
    void* address;
    void* table_of_contents;
    void* environment;
} FCN_DESC;

#endif _AIX

#ifdef hpux 

    /*
     *  Location of pc and sp in HP-UX jump buf
     */
#define HPPA_JB_PC(jmpbuf)   (((int *)(jmpbuf))[0])
#define HPPA_JB_SP(jmpbuf)   (((int *)(jmpbuf))[1])

#endif hpux


/*
 *	Note on passing arguments to forked functions:
 *		When a function is called by forking a thread to execute
 *		it, the function must realize it has been forked if any
 *		arguments are passed.  These arguments are stored in the
 *		args field of the TCB for all machines except vax's and
 *		mips.  All arguments must fit in a long.  This makes the thread 
 *		code more portable, at the expense of forking.
 *
 *		On mips boxes the args are stored in registers by placing them
 *		in the jmpbuf.  The jmpbug has room for 4 arguments stored in
 *		JB_A0 (JB_REGS+4) through JB_A3 (JB_REGS+7).
 */
#define	MAX_FORK_ARGS 4
typedef long FORKARGLIST[MAX_FORK_ARGS];

/*
 *	this structure holds the state of a thread
 */
typedef struct {

	FOUR 		*sp;
	FOUR		*stack;
	LISTELEMENT	controlList;
#ifdef vax 
	long		*lastFP;
	long		*lastSP;
	long		*lastAP;
#endif vax
#if defined(mips) || defined (sparc) || defined(_AIX) || defined(hpux)
	jmp_buf		initBuf;
	jmp_buf		currentBuf;
#endif


/*
 *	Arguments for forked functions, for all arch's but mips and vax
 */
#if defined(sparc) || defined(_AIX) || defined(hpux)
	FORKARGLIST args;
#endif

	int			id;
	int			error;
	int			errno;
	int			state;
	DISKMSG		*diskMessage;
	MESSAGE		message;
	IOVECTOR	iovec;
	int			linkId;
	LISTELEMENT	transList;
	char		*transRec;
	int			semCount;		/* debugging - # sems held by this thread	*/
	int			switchCount;	/* used by thread macros for 
									checking syncronization */
	SOCKADDR	replyAddr;  	/* for distr trans - UDP address of "peer" 	*/
    int         ipsid;          /* support for IPS performance monitor 		*/
                                /* not actually needed unless IPS_SUPPORT	*/
								/* is defined 								*/

	MAGIC		magic;
} TCB;



#define TCB_MAGIC	0x45728ddf


#if MAGIC_CHECKING IS_ENABLED


#define CHECK_TCB_MAGIC(_tcb)					\
												\
	if (_tcb->magic != TCB_MAGIC)	{			\
												\
		SM_ERROR(TYPE_FATAL, esmINTERNAL);		\
	}

#define INIT_TCB_MAGIC(_tcb)					\
												\
	_tcb->magic = TCB_MAGIC;


#else


#define CHECK_TCB_MAGIC(_tcb)

#define INIT_TCB_MAGIC(_tcb)


#endif


/*
 *	define the function prototypes
 */
#ifdef __cplusplus


extern "C" void		notify (LIST *, int, int);
extern "C" int		dispatch (int);


#else 


extern void		notify (LIST *, int, int);
extern int		dispatch (int);


#endif


/*
 *	define parent and child values
 */
#define T_CHILD		0x1
#define T_PARENT	0x2


/*
 *	These macros are used to check that thread switches have not
 *	occured in an interval of code.  This is done by calling
 *	STORE_THREAD_SWITCH at the beginning of the segment and calling
 *	CHECK_THREAD_SWITCH at the end.
 *
 *	ThreadSwitchCount is incremented in threadSwitch(). 
 */
#define STORE_THREAD_SWITCH(_tcb)	\
	(_tcb)->switchCount = ThreadSwitchCount

#define CHECK_THREAD_SWITCH(_tcb)	\
	SM_ASSERT(LEVEL_1, (_tcb)->switchCount == ThreadSwitchCount)


#endif __THREAD_H__
