#ifdef MEDIA_RECOVERY
/*
 *   $RCSfile: com_packet.c,v $  
 *   $Revision: 1.16 $  
 *   $Date: 1993/02/04 20:30:06 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
/*
 *
 * NOTE:
 * ----
 * The routines in this file assume that the structures that are being passed 
 * are in network order with respect to the individual fields of the substruc-
 * tures.  Therefore the packet which is composed of some structures like PID
 * LSN, etc should have the fields of PID, LSN, etc. converted to network order
 * before passing the packet to this routine.
 *
 */

#include "ess.h"
#include "checking.h"
#include "sysdefs.h"
#include "error.h"
#include "list.h"
#include "latch.h"
#include "io.h"
#include "tid.h"
#include "lock.h"
#include "msgdefs.h"
#include "thread.h"
#include "bf.h"
#include "com_packet.h"
#include "server_stub.h"
#include "mr_intfuncs.h"

#define COM_PACKET_DEBUG 0
#define COM_PACKET_DEBUG1 0

extern int errno;

/*
 * alloc_packet(size)
 *
 * This routine will allocate the given number of bytes by calling malloc().
 *
 */

char *alloc_packet(int size)
{

    char *newMem;

    newMem = (char*) malloc( size ); 

    if (newMem == NULL) {
		SM_ERROR(TYPE_FATAL, esmMALLOCFAILED);
    }
    return newMem;
}

/*
 * send_packet(sock, pstruct)
 *
 * This routine will construct a packet from the info given in the packet
 * structure and send it on the given socket.
 *
 * Return values: esmNOERROR for success. esmFAILURE for failure (with errno 
 * set to indicate the error)
 *
 */

int send_packet(int sock, packet* pstruct)
{
    int size;
    char *pack;
	int bytes_sent;
	int total_bytes;
	int remain_bytes;

                           /* alloc pack & fill in sequential fields */
    switch (pstruct->ptype) {   

        case RQ_LOG_VOL_HDR:
		    size = sizeof(rqlvh_pkt)+sizeof(pstruct->ptype);
		    pack = alloc_packet(size);
		    *pack = RQ_LOG_VOL_HDR;
		    BCOPY(&(pstruct->upacket.rqlvh), (pack+sizeof(pstruct->ptype)), 
														sizeof(rqlvh_pkt));
		    break;

        case RE_LOG_VOL_HDR:
		    size = sizeof(relvh_pkt)+sizeof(pstruct->ptype);
		    pack = alloc_packet(size);
		    *pack = RE_LOG_VOL_HDR;
		    BCOPY(&(pstruct->upacket.relvh), (pack+sizeof(pstruct->ptype)), 
													sizeof(relvh_pkt));
		    break;

        case INIT_DUMP:
		    size = sizeof(initdump_pkt)+sizeof(pstruct->ptype);
		    pack = alloc_packet(size);
		    *pack = INIT_DUMP;
		    BCOPY(&(pstruct->upacket.initdump), (pack+sizeof(pstruct->ptype)), 
														sizeof(initdump_pkt));
		    break;

        case MR_RQ_DATA_PAGE:
		    size = sizeof(mrrqdp_pkt)+sizeof(pstruct->ptype);
		    pack = alloc_packet(size);
		    *pack = MR_RQ_DATA_PAGE;
		    BCOPY(&(pstruct->upacket.mrrqdp), (pack+sizeof(pstruct->ptype)), 
														sizeof(mrrqdp_pkt));
		    break;

		case DUMP_DONE:
			size = sizeof(dumpdone_pkt) + sizeof(pstruct->ptype);
			pack = alloc_packet(size);
			*pack = DUMP_DONE;
			BCOPY(&(pstruct->upacket.dumpdone), (pack+sizeof(pstruct->ptype)),
														sizeof(dumpdone_pkt));
			break;

		case SR_RQ_DATA_PAGE:
		    size = sizeof(srrqdp_pkt)+sizeof(pstruct->ptype);
		    pack = alloc_packet(size);
		    *pack = SR_RQ_DATA_PAGE;
		    BCOPY(&(pstruct->upacket.srrqdp), (pack+sizeof(pstruct->ptype)), 
														sizeof(srrqdp_pkt));
	
		    break;

        case SR_RQ_LOG_PAGE:
		    size = sizeof(srrqlp_pkt)+sizeof(pstruct->ptype);
		    pack = alloc_packet(size);
		    *pack = SR_RQ_LOG_PAGE;
		    BCOPY(&(pstruct->upacket.srrqlp), (pack+sizeof(pstruct->ptype)), 
														sizeof(srrqlp_pkt));
		    break;
	
		case ACK:
			size = sizeof(ack_pkt)+sizeof(pstruct->ptype);
	   		pack = alloc_packet(size);
			*pack = ACK;
			BCOPY(&(pstruct->upacket.ack), (pack+sizeof(pstruct->ptype)), 
														sizeof(ack_pkt));
			break;

        /* SPECIAL CASES ... pointers */

		case SR_RE_DATA_PAGE:
		{
			int temp;
			size = (int) sizeof(pstruct->ptype) +
					sizeof(pstruct->upacket.srredp.pid) +
					sizeof(pstruct->upacket.srredp.length) +
					pstruct->upacket.srredp.length;

			pack = alloc_packet(size);
			*pack = SR_RE_DATA_PAGE;

			BCOPY((char *)&(pstruct->upacket.srredp.pid), (pack + 
				sizeof(pstruct->ptype)), sizeof(pstruct->upacket.srredp.pid));

			if (COM_PACKET_DEBUG1) {
				fprintf(stdout,"Length before sending = %d\n",
									pstruct->upacket.srredp.length);
			}
			temp = (int) htonl(pstruct->upacket.srredp.length);

			BCOPY(&temp, (pack + sizeof(pstruct->ptype) + 
						sizeof(pstruct->upacket.srredp.pid)), sizeof(temp));

			BCOPY(pstruct->upacket.srredp.page, (pack +sizeof(pstruct->ptype) +
								sizeof(pstruct->upacket.srredp.pid) + 
								sizeof(pstruct->upacket.srredp.length)),
								pstruct->upacket.srredp.length);
			break;
		}

		case MR_RE_DATA_PAGE:
		{
			int temp;

			size = sizeof(pstruct->ptype) +
					sizeof(pstruct->upacket.mrredp.pid) +
					sizeof(pstruct->upacket.mrredp.startLSN) +
					sizeof(pstruct->upacket.mrredp.length) +
					pstruct->upacket.mrredp.length;

			pack = alloc_packet(size);
			*pack = MR_RE_DATA_PAGE;

			BCOPY((char *)&(pstruct->upacket.mrredp.pid), (pack + 
				sizeof(pstruct->ptype)), sizeof(pstruct->upacket.mrredp.pid));

			BCOPY((char *)&(pstruct->upacket.mrredp.startLSN), (pack + 
				sizeof(pstruct->ptype)+ sizeof(pstruct->upacket.mrredp.pid)),
				sizeof(pstruct->upacket.mrredp.startLSN));


			temp = htonl(pstruct->upacket.mrredp.length);

			BCOPY((char *)&temp, (pack + sizeof(pstruct->ptype)+
								sizeof(pstruct->upacket.mrredp.pid) +
								sizeof(pstruct->upacket.mrredp.startLSN)),
					sizeof(pstruct->upacket.mrredp.length));

			BCOPY(pstruct->upacket.mrredp.page, (pack + sizeof(pstruct->ptype)+ 
				sizeof(pstruct->upacket.mrredp.pid) + 
				sizeof(pstruct->upacket.mrredp.startLSN)+
				sizeof(pstruct->upacket.mrredp.length)),
					pstruct->upacket.mrredp.length);

			break;
		}

		case DUMP_LOG_PAGE:
		{
			int temp;

			size = sizeof(pstruct->ptype) +
                    sizeof(pstruct->upacket.dumplp.length) +
					pstruct->upacket.dumplp.length;

			pack = alloc_packet(size);
			*pack = DUMP_LOG_PAGE;

			temp = htonl(pstruct->upacket.dumplp.length);

			BCOPY((char *)&temp, (pack + sizeof(pstruct->ptype)),
					sizeof(pstruct->upacket.dumplp.length));

			BCOPY(pstruct->upacket.dumplp.page, (pack + sizeof(pstruct->ptype)+ 
				sizeof(pstruct->upacket.dumplp.length)),
					pstruct->upacket.dumplp.length);

			break;
		}

		case MR_RE_LOG_PAGE:
		{
			int temp;

			size = sizeof(pstruct->ptype) +
                    sizeof(pstruct->upacket.mrrelp.pid) +
                    sizeof(pstruct->upacket.mrrelp.last_page) +
                    sizeof(pstruct->upacket.mrrelp.length) +
					pstruct->upacket.mrrelp.length;

			pack = alloc_packet(size);
			*pack = MR_RE_LOG_PAGE;

			BCOPY((char *)&(pstruct->upacket.mrrelp.pid), (pack + 
				sizeof(pstruct->ptype)), sizeof(pstruct->upacket.mrrelp.pid));

			BCOPY((char *)&(pstruct->upacket.mrrelp.last_page), (pack + 
					sizeof(pstruct->ptype) +
					sizeof(pstruct->upacket.mrrelp.pid)), 
				sizeof(pstruct->upacket.mrrelp.last_page));

			if (COM_PACKET_DEBUG1) {
				fprintf(stdout,"length of the page = %d\n", 
							pstruct->upacket.mrrelp.length);
			}

			temp = htonl(pstruct->upacket.mrrelp.length);

			BCOPY((char *)&temp, (pack + sizeof(pstruct->ptype)+
								sizeof(pstruct->upacket.mrrelp.last_page) +
								sizeof(pstruct->upacket.mrrelp.pid)),
					sizeof(pstruct->upacket.mrrelp.length));

			BCOPY(pstruct->upacket.mrrelp.page, (pack + sizeof(pstruct->ptype)+ 
				sizeof(pstruct->upacket.mrrelp.pid) + 
				sizeof(pstruct->upacket.mrrelp.last_page) +
				sizeof(pstruct->upacket.mrrelp.length)),
					pstruct->upacket.mrrelp.length);

			break;
		}

		default:
		    fprintf(stdout,"Invalid packet type %d given in send_packet()\n", 
															pstruct->ptype);
		    return esmFAILURE;

	}

	bytes_sent = 0;
	for (total_bytes = 0; total_bytes < size; total_bytes += bytes_sent)   {

		/* 
		 * Calculate remaining bytes to be sent 
		 */

		remain_bytes = size - total_bytes;

		if (COM_PACKET_DEBUG1) {
			 fprintf(stdout,"Sending %d bytes in this send\n",remain_bytes); 
		}

		/* 
		 * Send remaining bytes.
		 */

		if ((bytes_sent = send(sock, pack+total_bytes, remain_bytes, 0)) < 0) {	
			perror("send_packet");
			free(pack);
			return esmFAILURE;
		}
	}
	if (COM_PACKET_DEBUG1) {
		fprintf(stdout,"Bytes sent = %d\n",total_bytes); 
	}
	fflush(stdout);
	free(pack);
	return esmNOERROR;
}


/* ========================================================================= */

/*
 * rcv_packet(sock, pstruct)
 *
 * This routine will read the next packet from the socket given and insert the
 * pertinent info into the packet structure given.
 *
 * Return values: END_OF_FILE if socket has been torn down (i.e. the process at
 *                            the other end terminated)
 *                esmFAILURE for failure (with errno set to indicate the error)
 *		  a positive number for success.
 */

int rcv_packet(int sock, packet* pstruct)
{
	int ret;

	bzero((char *)pstruct, sizeof(packet));
	if (COM_PACKET_DEBUG1) {
		fprintf(stdout,"Calling rcv_packet_internal\n"); 
	}
	ret = rcv_packet_internal(sock, pstruct);
	if (COM_PACKET_DEBUG1) {
		fprintf(stdout,"Returned from rcv_packet_internal with ret = %d\n", ret); 
	}
	fflush(stdout);

	if (ret < 0)
		return(END_OF_FILE);
	else
		return esmNOERROR;
}


/* ========================================================================= */


int rcv_packet_internal(int sock, packet* pstruct)
{
    ONE temp;

	if (COM_PACKET_DEBUG1) {
		fprintf(stdout,"Calling recv_loop\n"); 
	}
    if (recv_loop(sock, &temp, sizeof(pstruct->ptype)) < 0) {
	perror("rcv_packet");
	return esmFAILURE;
    }
    pstruct->ptype = temp;
	if (COM_PACKET_DEBUG1) {
		fprintf(stdout,"Value of temp = %d\n", temp); 
		fprintf(stdout,"rcv_packet_internal: value of pstruct->ptype = %d\n",
														pstruct->ptype);
		fflush(stdout);
	}

    switch (pstruct->ptype) {

        case RQ_LOG_VOL_HDR:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a RQ_LOG_VOL_HDR\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.rqlvh), sizeof(rqlvh_pkt)));

		case RE_LOG_VOL_HDR:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a RE_LOG_VOL_HDR\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.relvh), sizeof(relvh_pkt)));

		case INIT_DUMP:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a INIT_DUMP\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.initdump), 
											sizeof(initdump_pkt)));

        case MR_RQ_DATA_PAGE:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a MR_RQ_DATA_PAGE\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.mrrqdp),
												sizeof(mrrqdp_pkt)));
		case DUMP_DONE:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a DUMP_DONE\n");
			}
			return(recv_loop(sock, &(pstruct->upacket.dumpdone),
										sizeof(pstruct->upacket.dumpdone)));

        case SR_RQ_DATA_PAGE:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a SR_RQ_DATA_PAGE\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.srrqdp),
												sizeof(srrqdp_pkt)));

        case SR_RQ_LOG_PAGE:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a SR_RQ_LOG_PAGE\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.srrqlp),
												sizeof(srrqlp_pkt)));

        case ACK:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a ACK\n");
			}
    	    return(recv_loop(sock, &(pstruct->upacket.ack), sizeof(ack_pkt)));




        /* SPECIAL CASES ... pointers */


        case SR_RE_DATA_PAGE:
		{
	    	int temp;
			int no_of_bytes;
			int total_no_of_bytes=1;

			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a SR_RE_DATA_PAGE\n");
			}
	        if ((no_of_bytes = recv_loop(sock, &(pstruct->upacket.srredp.pid), 
						sizeof(pstruct->upacket.srredp.pid))) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
			total_no_of_bytes += no_of_bytes;


	
			if ((no_of_bytes = recv_loop(sock, &temp, sizeof(temp))) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
			total_no_of_bytes += no_of_bytes;
	
			pstruct->upacket.srredp.length = ntohl(temp);
   	 	    
		    pstruct->upacket.srredp.page = 
							alloc_packet(pstruct->upacket.srredp.length + 1);
	
   		    if ((no_of_bytes = recv_loop(sock, pstruct->upacket.srredp.page, 
							pstruct->upacket.srredp.length)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
			total_no_of_bytes += no_of_bytes;
	
		    *(pstruct->upacket.srredp.page + pstruct->upacket.srredp.length) 
																		= '\0';
			if (COM_PACKET_DEBUG1) {
				fprintf(stdout,"Total No. of bytes in the packet = %d\n",
													total_no_of_bytes);
			}
		    return esmNOERROR;
		}

		case MR_RE_DATA_PAGE:
		{
			int temp;

			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a MR_RE_DATA_PAGE\n");
			}
	        if (recv_loop(sock, &(pstruct->upacket.mrredp.pid), 
						sizeof(pstruct->upacket.mrredp.pid)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
	        if (recv_loop(sock, &(pstruct->upacket.mrredp.startLSN), 
						sizeof(pstruct->upacket.mrredp.startLSN)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
			if (recv_loop(sock, &temp, sizeof(temp)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
			pstruct->upacket.mrredp.length = ntohl(temp);
   	 	    
		    pstruct->upacket.mrredp.page = 
							alloc_packet(pstruct->upacket.mrredp.length + 1);
	
   		    if (recv_loop(sock, pstruct->upacket.mrredp.page, 
							pstruct->upacket.mrredp.length) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
		    *(pstruct->upacket.mrredp.page + pstruct->upacket.mrredp.length) 
																		= '\0';
		    return esmNOERROR;
		}

        case DUMP_LOG_PAGE:
		{
	    	int temp;

			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a DUMP_LOG_PAGE\n");
			}
			if (recv_loop(sock, &temp, sizeof(temp)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
			pstruct->upacket.dumplp.length = ntohl(temp);
   	 	    
		    pstruct->upacket.dumplp.page = 
							alloc_packet(pstruct->upacket.dumplp.length + 1);
	
   		    if (recv_loop(sock, pstruct->upacket.dumplp.page, 
							pstruct->upacket.dumplp.length) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
		    *(pstruct->upacket.dumplp.page + pstruct->upacket.dumplp.length) 
																		= '\0';
		    return esmNOERROR;
		}


        case MR_RE_LOG_PAGE:
		{
	    	int temp;

			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a MR_RE_LOG_PAGE\n");
			}
	        if (recv_loop(sock, &(pstruct->upacket.mrrelp.pid), 
						sizeof(pstruct->upacket.mrrelp.pid)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
	        if (recv_loop(sock, &(pstruct->upacket.mrrelp.last_page), 
						sizeof(pstruct->upacket.mrrelp.last_page)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
			if (recv_loop(sock, &temp, sizeof(temp)) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
			pstruct->upacket.mrrelp.length = ntohl(temp);
			if (COM_PACKET_DEBUG1) {
				fprintf(stdout,"length = %d\n",pstruct->upacket.mrrelp.length); 
			}
   	 	    
		    pstruct->upacket.mrrelp.page = 
							alloc_packet(pstruct->upacket.mrrelp.length + 1);
	
   		    if (recv_loop(sock, pstruct->upacket.mrrelp.page, 
							pstruct->upacket.mrrelp.length) < 0) {
				perror("rcv_packet");
				return esmFAILURE;
   		    }
	
		    *(pstruct->upacket.mrrelp.page + pstruct->upacket.mrrelp.length) 
																		= '\0';
		    return esmNOERROR;
		}

		case END_OF_FILE:
			fprintf(stdout,"Received an END_OF_FILE on the socket\n");
			return esmFAILURE;

		default:
			if (COM_PACKET_DEBUG) {
				fprintf(stdout,"Received a Invalid packet\n");
			}
			return esmFAILURE;

	}
}

/* ========================================================================= */

int recv_loop(int sock, void* buffer, int read_bytes)
{
	int bytes_read;
	int remain_bytes;
	int total_bytes;

	/*
	 *  read the messages, repeatedly perhaps
	 */

	for (total_bytes = 0; total_bytes < read_bytes; 
											total_bytes += bytes_read)   {

		/*
		 *  calculate remaining bytes
		 */

		remain_bytes = read_bytes - total_bytes;

		/*
		 *  read the remaining bytes
		 */

		 bytes_read = recv(sock, ((char*)buffer) + total_bytes, remain_bytes, 0);

		if (COM_PACKET_DEBUG1) {
			fprintf(stdout,"bytes remaining = %d\n", remain_bytes);
			fprintf(stdout,"bytes read = %d\n", bytes_read);
		}

		 if (bytes_read <= 0) {
   	         perror("recv");
				return (END_OF_FILE);
		}
	}
	return total_bytes;
}


#endif MEDIA_RECOVERY
