/*
 *   $RCSfile: freeResource.c,v $  
 *   $Revision: 2.8 $  
 *   $Date: 1992/11/19 06:09:50 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "bitvec.h"
#include "msgvector.h"
#include "resource.h"
#include "resource_funcs.h"
#include "resource_globals.h"


#ifdef JUNK
#ifdef __cplusplus

extern "C" void		closeSequentialLoad (char *);
extern "C" void		closeSequentialScan (char *);

#else

extern void		closeSequentialLoad (char *);
extern void		closeSequentialScan (char *);

#endif
#endif JUNK


 void
freeResource (

	register RESOURCE	*resource 
)
{

	TRACE(TR_MSG, TR_LEVEL_1);

	/*
	 *	free the resource based on the type of resource
	 */
	switch (resource->type)	{

#ifdef JUNK
		case SEQUENTIALLOAD:

			closeSequentialLoad(resource->pointer);
			break;

		case SEQUENTIALSCAN:

			closeSequentialScan(resource->pointer);
			break;
#endif JUNK

		default:

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
	}

	/*
	 *	nullify the resource
	 */
	resource->unique = 0;
	resource->flags  = RESOURCE_FREE;

	/*
	 *	attempt to get a resource from the free list
	 */
	listMoveEnq( &FreeResources, &(resource->linkList) );
}
