/*
 *   $RCSfile: dispatch.c,v $  
 *   $Revision: 2.9 $  
 *   $Date: 1992/11/19 06:10:18 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "bitvec.h"
#include "link.h"
#include "msgvector.h"
#include "thread_funcs.h"
#include "thread_globals.h"
#include "msg_globals.h"


 int
dispatch (

	int				threadState 
)
{

	register TCB	*tcb;


	TRPRINT(TR_THREAD, TR_LEVEL_1, ("threadState:%d", threadState));

	/* attempt to get a thread off the ready queue */
	if ((tcb = (TCB *) listDeq( &ReadyList )) == NULL)	{

		TRPRINT(TR_THREAD, TR_LEVEL_2, ("ready list empty"));

		/*
		 *	get a thread off the free list
		 */
		if ((tcb = (TCB *) listDeq( &FreeList )) == NULL)	{

			TRPRINT(TR_THREAD, TR_LEVEL_2, ("getting spare thread"));

			/*
			 *	check to see if there is a spare thread
			 */
			if (tcb = Spare)	{

				TRPRINT(TR_THREAD, TR_LEVEL_2, ("using spare:%d", tcb->id));

				/*
				 *	mark the spare as used
				 */
				Spare = NULL;
				SelectMask = DiskMask;
				SelectRestrict = TRUE;

			} else {

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
		}
	}
	TRPRINT(TR_THREAD, TR_LEVEL_2, ("dispatching thread:%d", tcb->id));

	/*
	 *	save the state of the current thread
	 */
	Active->state = threadState;

	/*
	 *	jump to the new routine
	 */
	threadSwitch(tcb);
	TRPRINT(TR_THREAD, TR_LEVEL_2, ("thread:%d returned from longjmp", Active->id));

	/*
	 *	return the error code
	 */
	return(Active->error);
}
